/*
 * QLogic Fibre Channel HBA Driver
 * Copyright (c)  2003-2012 QLogic Corporation
 *
 * See LICENSE.qla2xxx for copyright and licensing details.
 */
#include "qla_def.h"
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/vmalloc.h>
#include <scsi/scsi_tcq.h>
#include <linux/utsname.h>

static char *
qla_skip_spaces(const char *str)
{
        while (qla_isspace(*str))
                ++str;
        return (char *)str;
}

static char *
qla_strim(char *s)
{
        size_t size;
        char *end;

        size = strlen(s);
        if (!size)
                return s;

         end = s + size - 1;
        while (end >= s && qla_isspace(*end))
                end--;
        *(end + 1) = '\0';

         return qla_skip_spaces(s);
}

/**
 * qlafx00_pci_config() - Setup ISPFX00 PCI configuration registers.
 * @ha: HA context
 *
 * Returns 0 on success.
 */
int
qlafx00_pci_config(scsi_qla_host_t *ha)
{
	uint16_t w;

	pci_set_master(ha->pdev);
	pci_try_set_mwi(ha->pdev);

	pci_read_config_word(ha->pdev, PCI_COMMAND, &w);
	w |= (PCI_COMMAND_PARITY | PCI_COMMAND_SERR);
	w &= ~PCI_COMMAND_INTX_DISABLE;
	pci_write_config_word(ha->pdev, PCI_COMMAND, w);

	/* PCIe -- adjust Maximum Read Request Size (2048). */
	if (pci_find_capability(ha->pdev, PCI_CAP_ID_EXP))
		pcie_set_readrq(ha->pdev, 2048);

	ha->chip_revision = ha->pdev->revision;

	return QLA_SUCCESS;
}

/**
 * qlafx00_warm_reset() - Perform warm reset of iSA(CPUs being reset on SOC).
 * @ha: HA context
 *
  */
static inline void
qlafx00_soc_cpu_reset(scsi_qla_host_t *ha)
{
	unsigned long flags = 0;
	int i, core;
	uint32_t cnt;

	/* Set all 4 cores in reset */
	for (i = 0; i < 4; i++ ){
		QLAFX00_SET_HBA_SOC_REG(ha,
		    (SOC_SW_RST_CONTROL_REG_CORE0 + 8*i), (0xF01));
	}

	/* Set all 4 core Clock gating control */
	for (i = 0; i < 4; i++ ){
		QLAFX00_SET_HBA_SOC_REG(ha,
		    (SOC_SW_RST_CONTROL_REG_CORE0 + 4 + 8*i), (0x01010101));
	}

	/* Reset all units in Fabric */
	QLAFX00_SET_HBA_SOC_REG(ha, SOC_FABRIC_RST_CONTROL_REG, (0x11F0101));

	/* Reset all interrupt control registers */
	for (i = 0; i < 115; i++ ){
		QLAFX00_SET_HBA_SOC_REG(ha,
		    (SOC_INTERRUPT_SOURCE_I_CONTROL_REG + 4*i), (0x0));
	}

	/* Reset Timers control registers. per core */
	for(core = 0; core < 4; core++)
		for(i = 0; i < 8; i++)
			QLAFX00_SET_HBA_SOC_REG(ha,
			    (SOC_CORE_TIMER_REG + 0x100*core + 4*i), (0x0));

	/* Reset per core IRQ ack register */
	for(core = 0; core < 4; core++)
		QLAFX00_SET_HBA_SOC_REG(ha,
		    (SOC_IRQ_ACK_REG + 0x100*core), (0x3FF));

	/* Set Fabric control and config to defaults */
	QLAFX00_SET_HBA_SOC_REG(ha, SOC_FABRIC_CONTROL_REG, (0x2));
	QLAFX00_SET_HBA_SOC_REG(ha, SOC_FABRIC_CONFIG_REG, (0x3));

	spin_lock_irqsave(&ha->hardware_lock, flags);

	/* Kick in Fabric units */
	QLAFX00_SET_HBA_SOC_REG(ha, SOC_FABRIC_RST_CONTROL_REG, (0x0));

	/* Kick in Core0 to start boot process */
	QLAFX00_SET_HBA_SOC_REG(ha, SOC_SW_RST_CONTROL_REG_CORE0, (0xF00));

	/* Wait for soft-reset to complete. */
	for (cnt = 150000; cnt; cnt--) {
		udelay(100);
		barrier();
	}
	spin_unlock_irqrestore(&ha->hardware_lock, flags);
}

/**
 * qlafx00_soft_reset() - Soft Reset ISPFx00.
 * @ha: HA context
 *
 * Returns 0 on success.
 */
void
qlafx00_soft_reset(scsi_qla_host_t *ha)
{

	if (unlikely(pci_channel_offline(ha->pdev) &&
	    ha->flags.pci_channel_io_perm_failure))
		return;

	ha->isp_ops->disable_intrs(ha);
	qlafx00_soc_cpu_reset(ha);
	ha->isp_ops->enable_intrs(ha);
}

/**
 * qlafx00_chip_diag() - Test ISPFX00 for proper operation.
 * @ha: HA context
 *
 * Returns 0 on success.
 */
int
qlafx00_chip_diag(scsi_qla_host_t *ha)
{
	int rval = 0;

	ha->fw_transfer_size = cpu_to_le16(ha->request_q_length);

	rval = qlafx00_mbx_reg_test(ha);
	if (rval) {
		qla_printk(KERN_WARNING, ha,
		    "Failed mailbox send register test\n");
	} else {
		/* Flag a successful rval */
		rval = QLA_SUCCESS;
	}
	return rval;
}

void
qlafx00_config_rings(struct scsi_qla_host *ha)
{
	struct device_reg_fx00 __iomem *reg = &ha->iobase->ispfx00;
	init_cb_fx_t *icb;

	/* Setup ring parameters in initialization control block. */
	icb = (init_cb_fx_t *)ha->init_cb;
	icb->request_q_outpointer = __constant_cpu_to_le16(0);
	icb->response_q_inpointer = __constant_cpu_to_le16(0);
	icb->request_q_length = cpu_to_le16(ha->request_q_length);
	icb->response_q_length = cpu_to_le16(ha->response_q_length);
	icb->request_q_address[0] = cpu_to_le32(LSD(ha->request_dma));
	icb->request_q_address[1] = cpu_to_le32(MSD(ha->request_dma));
	icb->response_q_address[0] = cpu_to_le32(LSD(ha->response_dma));
	icb->response_q_address[1] = cpu_to_le32(MSD(ha->response_dma)); 

	WRT_REG_DWORD(&reg->req_q_in, 0);
	WRT_REG_DWORD(&reg->req_q_out, 0);

	WRT_REG_DWORD(&reg->rsp_q_in, 0);
	WRT_REG_DWORD(&reg->rsp_q_out, 0);

	/* PCI posting */
	RD_REG_DWORD(&reg->rsp_q_out);
}

char *
qlafx00_pci_info_str(struct scsi_qla_host *ha, char *str)
{
	int pcie_reg;

	pcie_reg = pci_find_capability(ha->pdev, PCI_CAP_ID_EXP);
	if (pcie_reg) {
		strcpy(str, "PCIe iSA");
		return str;
	}
	return str;
}

char *
qlafx00_fw_version_str(struct scsi_qla_host *ha, char *str)
{

	sprintf(str, "%s", ha->mr.fw_version);
	return str;
}

void
qlafx00_enable_intrs(struct scsi_qla_host *ha)
{
	unsigned long flags = 0;

	spin_lock_irqsave(&ha->hardware_lock, flags);
	ha->interrupts_on = 1;
	QLAFX00_ENABLE_ICNTRL_REG(ha);
	spin_unlock_irqrestore(&ha->hardware_lock, flags);
}

void
qlafx00_disable_intrs(struct scsi_qla_host *ha)
{
	unsigned long flags = 0;

	spin_lock_irqsave(&ha->hardware_lock, flags);
	ha->interrupts_on = 0;
	QLAFX00_DISABLE_ICNTRL_REG(ha);
	spin_unlock_irqrestore(&ha->hardware_lock, flags);
}

void
qlafx00_wait_for_abort_completion(scsi_qla_host_t *ha, struct abt *pabt)
{
	unsigned long timeout;

	if (unlikely(pci_channel_offline(ha->pdev)))
		return;

	timeout = ((ha->r_a_tov / 10 * 2) + 20) * HZ;
	if (!wait_for_completion_timeout(&pabt->abt_comp, timeout)) {
		DEBUG2(qla_printk(KERN_WARNING, ha,
		    "ABT IOCB timeout.\n"));
	}
}

int
qlafx00_abort_command(scsi_qla_host_t *ha, srb_t *sp)
{
	unsigned long   flags = 0;
	uint32_t	handle;

	spin_lock_irqsave(&ha->hardware_lock, flags);
	for (handle = 1; handle < MAX_OUTSTANDING_COMMANDS; handle++) {
		if (ha->outstanding_cmds[handle] == sp)
			break;
	}
	spin_unlock_irqrestore(&ha->hardware_lock, flags);
	if (handle == MAX_OUTSTANDING_COMMANDS) {
		/* Command not found. */
		return QLA_FUNCTION_FAILED;
	}
	return qlafx00_async_abt_cmd(handle, sp);
}

void
qlafx00_wait_for_tmf_completion(scsi_qla_host_t *ha, struct tmf *ptmf)
{
	unsigned long timeout;

	if (unlikely(pci_channel_offline(ha->pdev)))
		return;

	timeout = ((ha->r_a_tov / 10 * 2) + 20) * HZ;
	if (!wait_for_completion_timeout(&ptmf->tmf_comp, timeout)) {
		DEBUG2(qla_printk(KERN_WARNING, ha,
		    "TMF IOCB timeout.\n"));
	}
}

int qlafx00_async_tm_cmd(fc_port_t *fcport, uint32_t flags,
    uint32_t lun)
{
	scsi_qla_host_t *ha = fcport->ha;
	struct tsk_mgmt_entry_fx00 tmf_iocb = {0};
	struct tmf taskm;
	unsigned long itr_flags;
	uint32_t rval = CS_TIMEOUT;

	DEBUG19(qla_printk(KERN_INFO, ha,
	    "Task Management command issued, flags=0x%x\n", flags));

	taskm.flags = flags;
	taskm.lun = lun;
	init_completion(&taskm.tmf_comp);

	qlafx00_tm_iocb(fcport, &tmf_iocb, &taskm);
	spin_lock_irqsave(&ha->hardware_lock, itr_flags);
	qla2x00_isp_cmd(ha);
	spin_unlock_irqrestore(&ha->hardware_lock, itr_flags);

	qlafx00_wait_for_tmf_completion(ha, &taskm);

	rval = taskm.comp_status == CS_COMPLETE ?
	    QLA_SUCCESS: QLA_FUNCTION_FAILED;

	return rval;
}

int qlafx00_abort_target(fc_port_t *fcport)
{
	return qlafx00_async_tm_cmd(fcport, TCF_TARGET_RESET, 0);
}

int
qlafx00_loop_reset(scsi_qla_host_t *ha) 
{
	int ret;
	struct fc_port *fcport;

	if (ql2xtargetreset) {
		list_for_each_entry(fcport, &ha->fcports, list) {
			if (fcport->port_type != FCT_TARGET ||
			    ha->vp_idx != fcport->vp_idx)
				continue;

			ret = qla2x00_device_reset(ha, fcport);
			if (ret != QLA_SUCCESS) {
				DEBUG2_3(printk("%s(%ld): bus_reset failed: "
				    "target_reset=%d d_id=%x.\n", __func__,
				    ha->host_no, ret, fcport->d_id.b24));
			}
		}
	}
	return QLA_SUCCESS;
}

int
qlafx00_iospace_config(struct scsi_qla_host *ha)
{
	if (pci_request_regions(ha->pdev, QLA2XXX_DRIVER_NAME)) {
		qla_printk(KERN_ERR, ha,
		    "Failed to reserve PIO/MMIO regions (%s), aborting.\n",
		    pci_name(ha->pdev));
		goto iospace_error_exit;
	}

	/* Use MMIO operations for all accesses. */
	if (!(pci_resource_flags(ha->pdev, 0) & IORESOURCE_MEM)) {
		qla_printk(KERN_WARNING, ha,
		    "Invalid pci I/O region size (%s).\n",
		    pci_name(ha->pdev));
		goto iospace_error_exit;
	}
	if (pci_resource_len(ha->pdev, 0) < BAR0_LEN_FX00) {
		qla_printk(KERN_WARNING, ha,
		    "Invalid PCI mem BAR0 region size (%s), aborting\n",
			pci_name(ha->pdev));
		goto iospace_error_exit;
	}

	ha->cregbase = ioremap_nocache(pci_resource_start(ha->pdev, 0), BAR0_LEN_FX00);
	if (!ha->cregbase) {
		qla_printk(KERN_ERR, ha,
		    "cannot remap MMIO (%s), aborting\n", pci_name(ha->pdev));
		goto iospace_error_exit;
	}

	if (!(pci_resource_flags(ha->pdev, 2) & IORESOURCE_MEM)) {
		qla_printk(KERN_WARNING, ha,
		    "region #2 not an MMIO resource (%s), aborting\n",
		    pci_name(ha->pdev));
		goto iospace_error_exit;
	}
	if (pci_resource_len(ha->pdev, 2) < BAR2_LEN_FX00) {
		qla_printk(KERN_WARNING, ha,
		    "Invalid PCI mem BAR2 region size (%s), aborting\n",
			pci_name(ha->pdev));
		goto iospace_error_exit;
	}

	ha->iobase = ioremap_nocache(pci_resource_start(ha->pdev, 2), BAR2_LEN_FX00);
	if (!ha->iobase) {
		qla_printk(KERN_ERR, ha,
		    "cannot remap MMIO (%s), aborting\n", pci_name(ha->pdev));
		goto iospace_error_exit;
	}

	/* Determine queue resources */

	qla_printk(KERN_INFO, ha,
	    "Iobase0 0x%p, iobase2 0x%p\n",
	    ha->cregbase, ha->iobase);

	return (0);

iospace_error_exit:
	return (-ENOMEM);
}
void
qlafx00_save_queue_ptrs(struct scsi_qla_host *ha)
{

	ha->fx00_request_q_length = ha->request_q_length;
	ha->fx00_request_ring = ha->request_ring;
	ha->fx00_request_dma = ha->request_dma;

	ha->fx00_response_q_length = ha->response_q_length;
	ha->fx00_response_ring = ha->response_ring;
	ha->fx00_response_dma = ha->response_dma;

	DEBUG19(qla_printk(KERN_INFO, ha,
	    "fx00_request_ring: %p, fx00_request_dma: 0x%llx "
	    "fx00_request_q_length: 0x%x\n",
	    ha->fx00_request_ring, ha->fx00_request_dma,
	    ha->fx00_request_q_length));


	DEBUG19(qla_printk(KERN_INFO, ha,
	    "fx00_response_ring: %p, fx00_response_dma: 0x%llx "
	    "fx00_response_q_length: 0x%x\n",
	    ha->fx00_response_ring, ha->fx00_response_dma,
	    ha->fx00_response_q_length));
}

int
qlafx00_config_queues(struct scsi_qla_host *ha)
{
	dma_addr_t bar2_hdl = pci_resource_start(ha->pdev, 2);

	ha->request_q_length = ha->req_que_len;
	ha->request_ring = (void *)ha->iobase + ha->req_que_off;
	ha->request_dma = bar2_hdl + ha->req_que_off;
	if ((!ha->request_ring) || (ha->request_q_length == 0)) {
		qla_printk(KERN_INFO, ha,
		    "Unable to allocate memory for req_ring\n");
		return QLA_FUNCTION_FAILED;
	}

	DEBUG19(qla_printk(KERN_INFO, ha,
	    "request_ring: %p, request_dma: 0x%llx "
	    "request_q_length: 0x%x\n",
	    ha->request_ring, (u64)ha->request_dma,
	    ha->request_q_length));

	ha->response_q_length = ha->rsp_que_len;
	ha->response_ring = (void *)ha->iobase + ha->rsp_que_off;
	ha->response_dma = bar2_hdl + ha->rsp_que_off;
	if ((!ha->response_ring) || (ha->response_q_length == 0)) {
		qla_printk(KERN_INFO, ha,
		    "Unable to allocate memory for rsp_ring\n");
		return QLA_FUNCTION_FAILED;
	}
	DEBUG19(qla_printk(KERN_INFO, ha,
	    "response_ring: %p, response_dma: 0x%llx "
	    "response_q_length: 0x%x\n",
	    ha->response_ring, (u64)ha->response_dma,
	    ha->response_q_length));


	return QLA_SUCCESS;
}

int
qlafx00_init_fw_ready(scsi_qla_host_t *ha)
{
	int rval = 0;
	unsigned long wtime;
	uint16_t wait_time;	/* Wait time */
	struct device_reg_fx00 __iomem *reg = &ha->iobase->ispfx00;
	uint32_t aenmbx, aenmbx7 = 0;
	uint32_t pseudo_aen;
	uint32_t state[5];
	bool done = false;

	/* 30 seconds wait - Adjust if required */
	wait_time = 30;

	pseudo_aen = RD_REG_DWORD(&reg->pseudoaen);
	if (pseudo_aen == 1) {
		aenmbx7 = RD_REG_DWORD(&reg->initval7);
		ha->mbx_intr_code = MSW(aenmbx7);
		ha->rqstq_intr_code = LSW(aenmbx7);
		rval = qlafx00_driver_shutdown(ha, 10);
		if (rval != QLA_SUCCESS)
			qlafx00_soft_reset(ha);
	}
	/* wait time before firmware ready */
	wtime = jiffies + (wait_time * HZ);
	do {
		aenmbx = RD_REG_DWORD(&reg->aenmailbox0);
		barrier();
		DEBUG19(qla_printk(KERN_INFO, ha,
		    "aenmbx: 0x%x\n", aenmbx));

		switch (aenmbx) {
		case MBA_FW_NOT_STARTED:
		case MBA_FW_STARTING:
			break;

		case MBA_SYSTEM_ERR:
		case MBA_REQ_TRANSFER_ERR:
		case MBA_RSP_TRANSFER_ERR:
		case MBA_FW_INIT_FAILURE:
			qlafx00_soft_reset(ha);
			break;

		case MBA_FW_RESTART_CMPLT:
			/* Set the mbx and rqstq intr code */
			aenmbx7 = RD_REG_DWORD(&reg->aenmailbox7);
			ha->mbx_intr_code = MSW(aenmbx7);
			ha->rqstq_intr_code = LSW(aenmbx7);
			ha->req_que_off = RD_REG_DWORD(&reg->aenmailbox1);
			ha->rsp_que_off = RD_REG_DWORD(&reg->aenmailbox3);
			ha->req_que_len = RD_REG_DWORD(&reg->aenmailbox5);
			ha->rsp_que_len = RD_REG_DWORD(&reg->aenmailbox6);
			WRT_REG_DWORD(&reg->aenmailbox0, 0);
			RD_REG_DWORD_RELAXED(&reg->aenmailbox0);
			DEBUG19(qla_printk(KERN_INFO, ha,
			    "f/w returned mbx_intr_code: 0x%x, rqstq_intr_code: 0x%x\n",
			    ha->mbx_intr_code, ha->rqstq_intr_code));
			QLAFX00_CLR_INTR_REG(ha, QLAFX00_HST_INT_STS_BITS);
			rval = QLA_SUCCESS;
			done = true;
			break;

		default:
			/* If fw is apparently not ready. In order to continue, we
			 * might need to issue Mbox cmd, but the problem is that the
			 * DoorBell vector values that come with the 8060 AEN are
			 * most likely gone by now (and thus no bell would be rung
			 * on the fw side when mbox cmd is issued). We have to
			 * therefore grab the 8060 AEN shadow regs (filled in by FW
			 * when the last 8060 AEN was being posted).
			 * Do the following to determine what is needed in order to
			 * get the FW ready:
			 * 1. reload the 8060 AEN values from the shadow regs
			 * 2. clear int status to get rid of possible pending ints
			 * 3. issue Get FW State Mbox cmd to determine fw state
			 * Set the mbx and rqstq intr code from Shadow Regs
			 */
			aenmbx7 = RD_REG_DWORD(&reg->initval7);
			ha->mbx_intr_code = MSW(aenmbx7);
			ha->rqstq_intr_code = LSW(aenmbx7);
			ha->req_que_off = RD_REG_DWORD(&reg->initval1);
			ha->rsp_que_off = RD_REG_DWORD(&reg->initval3);
			ha->req_que_len = RD_REG_DWORD(&reg->initval5);
			ha->rsp_que_len = RD_REG_DWORD(&reg->initval6);
			DEBUG19(qla_printk(KERN_INFO, ha,
			    "f/w returned mbx_intr_code: 0x%x, rqstq_intr_code: 0x%x\n",
			    ha->mbx_intr_code, ha->rqstq_intr_code));
			QLAFX00_CLR_INTR_REG(ha, QLAFX00_HST_INT_STS_BITS);

			/* Get the FW state */
			rval = qlafx00_get_firmware_state(ha, state);
			if (rval != QLA_SUCCESS) {
				/* Retry if timer has not expired */
				break;
			}

			if (state[0] == FSTATE_FX00_CONFIG_WAIT) {
				/* Firmware is waiting to be initialized by driver */
				rval = QLA_SUCCESS;
				done = true;
				break;
			}

			/* Issue driver shutdown and wait until f/w recovers.
			 * Driver should continue to poll until 8060 AEN is received
			 * indicatinf firmware recovery.
			 */
			DEBUG19(qla_printk(KERN_INFO, ha,
			    "Sending Driver shutdown fw_state 0x%x \n",
			    state[0]));

			rval = qlafx00_driver_shutdown(ha, 10);
			if (rval != QLA_SUCCESS) {
				rval = QLA_FUNCTION_FAILED;
				break;
			}
			msleep(500);

			wtime = jiffies + (wait_time * HZ);
			break;
		}

		if (!done) {
			if (time_after_eq(jiffies, wtime)) {
				DEBUG2(qla_printk(KERN_INFO, ha,
				    "Init f/w failed: aen[7]: 0x%x\n",
				    RD_REG_DWORD(&reg->aenmailbox7)));
				rval = QLA_FUNCTION_FAILED;
				done = true;
				break;
			}
			/* Delay for a while */
			msleep(500);
		}
	} while (!done);

	if (rval)
		qla_printk(KERN_INFO, ha,
		    "%s **** FAILED ****.\n", __func__);
	else
		DEBUG19(qla_printk(KERN_INFO, ha,
		    "%s **** SUCCESS ****.\n", __func__));

	return (rval);
}

/*
 * qlafx00_fw_ready() - Waits for firmware ready.
 * @ha: HA context
 *
 * Returns 0 on success.
 */
int
qlafx00_fw_ready(scsi_qla_host_t *ha)
{
	int		rval;
	unsigned long	wtime;
	uint16_t	wait_time;	/* Wait time if loop is coming ready */
	uint32_t	state[5];

	rval = QLA_SUCCESS;

	wait_time = 10;

	/* wait time before firmware ready */
	wtime = jiffies + (wait_time * HZ);

	/* Wait for ISP to finish init */
	if (!ha->flags.init_done)
		DEBUG2(qla_printk(KERN_INFO, ha,
		    "Waiting for init to complete...\n"));

	do {
		rval = qlafx00_get_firmware_state(ha, state);

		if(rval == QLA_SUCCESS) {
			if (state[0] == FSTATE_FX00_INITIALIZED) {
				DEBUG2(qla_printk(KERN_INFO, ha,
				    "fw_state=%x\n", state[0]));
				rval = QLA_SUCCESS;
				break;
			}
		}
		rval = QLA_FUNCTION_FAILED;

		if (time_after_eq(jiffies, wtime))
			break;

		/* Delay for a while */
		msleep(500);

		DEBUG2(qla_printk(KERN_INFO, ha,
		    "fw_state=%x curr time=%lx.\n", state[0], jiffies));
	} while (1);


	if (rval)
		qla_printk(KERN_INFO, ha,
		    "Firmware ready **** FAILED ****.\n");
	else
		DEBUG2(qla_printk(KERN_INFO, ha,
		    "Firmware ready **** SUCCESS ****.\n"));

	return (rval);
}

static int
qlafx00_find_all_targets(scsi_qla_host_t *ha,
	struct list_head *new_fcports)
{
	int		rval;
	uint16_t	tgt_id;
	fc_port_t	*fcport, *new_fcport;
	int		found;

	rval = QLA_SUCCESS;

	if (!test_bit(LOOP_RESYNC_ACTIVE, &ha->dpc_flags))
		return (QLA_FUNCTION_FAILED);

	if ((atomic_read(&ha->loop_down_timer) ||
	     STATE_TRANSITION(ha))) {
		atomic_set(&ha->loop_down_timer, 0);
		set_bit(LOOP_RESYNC_NEEDED, &ha->dpc_flags);
		return (QLA_FUNCTION_FAILED);
	}

	DEBUG19(qla_printk(KERN_INFO, ha,
	    "Listing Target bit map...\n"));
	DEBUG19(qla2x00_dump_buffer((uint8_t *)ha->gid_list, 32));

	/* Allocate temporary rmtport for any new rmtports discovered. */
	new_fcport = qla2x00_alloc_fcport(ha, GFP_KERNEL);
	if (new_fcport == NULL)
		return (QLA_MEMORY_ALLOC_FAILED);

	qla_for_each_bit(tgt_id, (void*)ha->gid_list,  QLAFX00_TGT_NODE_LIST_SIZE) {

		/* Send get target node info */
		new_fcport->tgt_id = tgt_id;
		rval = qlafx00_fx_disc(ha, new_fcport,
		    FXDISC_GET_TGT_NODE_INFO);
		if (rval != QLA_SUCCESS) {
			qla_printk(KERN_WARNING, ha,
			    "Target info scan failed -- assuming zero-entry "
			    "result...\n");
			continue;
		}

		/* Locate matching device in database. */
		found = 0;
		list_for_each_entry(fcport, &ha->vp_fcports, list) {
			if (memcmp(new_fcport->port_name,
			    fcport->port_name, WWN_SIZE))
				continue;

			found++;

			/*
			 * If tgt_id is same and state FCS_ONLINE, nothing
			 * changed.
			 */
			if (fcport->tgt_id == new_fcport->tgt_id &&
			    atomic_read(&fcport->state) == FCS_ONLINE)
				break;

			/*
			 * Tgt ID changed or device was marked to be updated.
			 */
			DEBUG2(qla_printk(KERN_INFO, ha,
			    "TGT-ID Change(%s): Present tgt id: 0x%x state: 0x%x "
			    "wwnn = %llx wwpn = %llx.\n",
			    __func__, fcport->tgt_id,
			    atomic_read(&fcport->state),
			    (unsigned long long)wwn_to_u64(fcport->node_name),
			    (unsigned long long)wwn_to_u64(fcport->port_name)));

			qla_printk(KERN_INFO, ha,
			    "TGT-ID Announce(%s): Discovered tgt id 0x%x wwnn = %llx "
			    "wwpn = %llx.\n", __func__, new_fcport->tgt_id,
			    (unsigned long long)wwn_to_u64(new_fcport->node_name),
			    (unsigned long long)wwn_to_u64(new_fcport->port_name));

			if (atomic_read(&fcport->state) != FCS_ONLINE) {
				fcport->old_tgt_id = fcport->tgt_id;
				fcport->tgt_id = new_fcport->tgt_id;
				qla_printk(KERN_INFO, ha,
				   "TGT-ID: New fcport Added: %p\n", fcport);
				qla2x00_update_fcport(ha, fcport);
			} else {
				qla_printk(KERN_INFO, ha,
				    " Existing TGT-ID %x did not get "
				    " offline event from firmware.\n",
				    fcport->old_tgt_id);
				qla2x00_mark_device_lost(ha, fcport, 0, 0);
				set_bit(LOOP_RESYNC_NEEDED, &ha->dpc_flags);
				kfree(new_fcport);
				return(rval);
			}
			break;
		}

		if (found)
			continue;

		/* If device was not in our fcports list, then add it. */
		list_add_tail(&new_fcport->list, new_fcports);

		/* Allocate a new replacement fcport. */
		new_fcport = qla2x00_alloc_fcport(ha, GFP_KERNEL);
		if (new_fcport == NULL)
			return (QLA_MEMORY_ALLOC_FAILED);
	}

	kfree(new_fcport);
	return (rval);
}

/*
 * qlafx00_configure_all_targets
 *      Setup target devices with node ID's.
 *
 * Input:
 *      ha = adapter block pointer.
 *
 * Returns:
 *      0 = success.
 *      BIT_0 = error
 */
int
qlafx00_configure_all_targets(scsi_qla_host_t *ha)
{
	int rval;
	fc_port_t *fcport, *rmptemp;
	LIST_HEAD(new_fcports);

	rval = qlafx00_fx_disc(ha, &ha->mr.fcport, FXDISC_GET_TGT_NODE_LIST);
	if (rval != QLA_SUCCESS) {
		set_bit(LOOP_RESYNC_NEEDED, &ha->dpc_flags);
		return (rval);
	}

	rval = qlafx00_find_all_targets(ha, &new_fcports);
	if (rval != QLA_SUCCESS) {
		set_bit(LOOP_RESYNC_NEEDED, &ha->dpc_flags);
		return (rval);
	}

	/*
	 * Delete all previous devices marked lost.
	 */
	list_for_each_entry(fcport, &ha->vp_fcports, list) {
		if (test_bit(LOOP_RESYNC_NEEDED, &ha->dpc_flags))
			break;

		if (atomic_read(&fcport->state) == FCS_DEVICE_LOST) {
			if (fcport->port_type != FCT_INITIATOR)
				qla2x00_mark_device_lost(ha, fcport, 0, 0);
		}
	}

	/*
	 * Add the new devices to our devices list.
	 */
	list_for_each_entry_safe(fcport, rmptemp, &new_fcports, list) {
		if (test_bit(LOOP_RESYNC_NEEDED, &ha->dpc_flags))
			break;

		qla2x00_update_fcport(ha, fcport);
		list_move_tail(&fcport->list, &ha->vp_fcports);
		qla_printk(KERN_INFO, ha,
		    "Attach new target id 0x%x wwnn = %llx "
		    "wwpn = %llx.\n",
		    fcport->tgt_id,
		    (unsigned long long)wwn_to_u64(fcport->node_name),
		    (unsigned long long)wwn_to_u64(fcport->port_name));
	}

	/* Free all new device structures not processed. */
	list_for_each_entry_safe(fcport, rmptemp, &new_fcports, list) {
		list_del(&fcport->list);
		kfree(fcport);
	}

	return (rval);
}

/*
 * qlafx00_configure_devices
 *      Updates Fibre Channel Device Database with what is actually on loop.
 *
 * Input:
 *      ha                = adapter block pointer.
 *
 * Returns:
 *      0 = success.
 *      1 = error.
 *      2 = database was full and device was not configured.
 */
int
qlafx00_configure_devices(scsi_qla_host_t *ha)
{
	int  rval;
	unsigned long flags, save_flags;
	rval = QLA_SUCCESS;

	save_flags = flags = ha->dpc_flags;

	rval = qlafx00_configure_all_targets(ha);

	if (rval == QLA_SUCCESS) {
		if (test_bit(LOOP_RESYNC_NEEDED, &ha->dpc_flags)) {
			rval = QLA_FUNCTION_FAILED;
		} else {
			atomic_set(&ha->loop_state, LOOP_READY);
			qla_printk(KERN_INFO, ha,
			    "Device Ready\n");
		}
	}

	if (rval) {
		DEBUG2(qla_printk(KERN_INFO, ha,
		    "%s *** FAILED ***.\n", __func__));
	} else {
		DEBUG2(qla_printk(KERN_INFO, ha,
		    "%s: exiting normally.\n", __func__));
	}
	return (rval);
}

void
qlafx00_abort_isp_cleanup(scsi_qla_host_t *ha, bool critemp)
{
	fc_port_t *fcport;

	ha->flags.online = 0;
	ha->mr.fw_hbt_en = 0;

	if (!critemp) {
		ha->flags.chip_reset_done = 0;
		clear_bit(ISP_ABORT_NEEDED, &ha->dpc_flags);
		ha->qla_stats.total_isp_aborts++;

		qla_printk(KERN_INFO, ha,
		    "Performing ISP error recovery - ha = %p.\n", ha);

		ha->isp_ops->reset_chip(ha);
	}

	if (atomic_read(&ha->loop_state) != LOOP_DOWN) {
		atomic_set(&ha->loop_state, LOOP_DOWN);
		atomic_set(&ha->loop_down_timer,
		    QLAFX00_LOOP_DOWN_TIME);
	} else {
		if (!atomic_read(&ha->loop_down_timer))
			atomic_set(&ha->loop_down_timer,
			    QLAFX00_LOOP_DOWN_TIME);
	}

	/* Clear all async request states across all VPs. */
	list_for_each_entry(fcport, &ha->vp_fcports, list) {
		fcport->flags = 0;
		if (atomic_read(&fcport->state) == FCS_ONLINE)
			atomic_set(&fcport->state, FCS_DEVICE_LOST);
	}

	if (!ha->flags.eeh_busy) {
		if(critemp) {
			qla2x00_abort_all_cmds(ha, DID_NO_CONNECT << 16);
		} else {
			/* Requeue all commands in outstanding command list. */
			qla2x00_abort_all_cmds(ha, DID_RESET << 16);
		}
	}

	qla2x00_free_irqs(ha);
	if (critemp)
		set_bit(FX00_CRITEMP_RECOVERY, &ha->dpc_flags);
	else
		set_bit(FX00_RESET_RECOVERY, &ha->dpc_flags);


	/* Clear the Interupts */
	QLAFX00_CLR_INTR_REG(ha, QLAFX00_HST_INT_STS_BITS);

	qla_printk(KERN_INFO, ha,
	    "%s Done done - ha=%p.\n", __func__, ha);

}

int
qlafx00_rescan_isp(scsi_qla_host_t *ha)
{
	uint32_t status = QLA_FUNCTION_FAILED;
	struct device_reg_fx00 __iomem *reg = &ha->iobase->ispfx00;
	uint32_t aenmbx7;

	qla2x00_request_irqs(ha);

	aenmbx7 = RD_REG_DWORD(&reg->aenmailbox7);
	ha->mbx_intr_code = MSW(aenmbx7);
	ha->rqstq_intr_code = LSW(aenmbx7);
	ha->req_que_off = RD_REG_DWORD(&reg->aenmailbox1);
	ha->rsp_que_off = RD_REG_DWORD(&reg->aenmailbox3);
	ha->req_que_len = RD_REG_DWORD(&reg->aenmailbox5);
	ha->rsp_que_len = RD_REG_DWORD(&reg->aenmailbox6);

	DEBUG2(qla_printk(KERN_INFO, ha,
	    "fw returned mbx_intr_code: 0x%x, rqstq_intr_code: 0x%x "
	    " Req que offset 0x%x Rsp que offset 0x%x\n",
	    ha->mbx_intr_code, ha->rqstq_intr_code,
	    ha->req_que_off, ha->rsp_que_len));

	/* Clear the Interupts */
	QLAFX00_CLR_INTR_REG(ha, QLAFX00_HST_INT_STS_BITS);

	status = qla2x00_init_rings(ha);
	if (!status) {
		ha->flags.online = 1;
		/* if no cable then assume it's good */
		if ((ha->device_flags & DFLG_NO_CABLE))
			status = 0;
		/* Register system information */
		if (qlafx00_fx_disc(ha, &ha->mr.fcport, FXDISC_REG_HOST_INFO))
			DEBUG2(qla_printk(KERN_INFO, ha,
			    "(%s) Failed to register host info.\n", __func__));
	}
	scsi_unblock_requests(ha->host);
	return status;
}

void
qlafx00_timer_routine(scsi_qla_host_t *ha)
{
	uint32_t fw_heart_beat;
	uint32_t aenmbx0;
	struct device_reg_fx00 __iomem *reg = &ha->iobase->ispfx00;
	uint32_t tempc;

	/* Check firmware health */
	if (ha->mr.fw_hbt_cnt)
		ha->mr.fw_hbt_cnt--;
	else {
		if ((!ha->flags.mr_reset_hdlr_active) &&
		    (!test_bit(UNLOADING, &ha->dpc_flags)) &&
		    (!test_bit(ABORT_ISP_ACTIVE, &ha->dpc_flags)) &&
		    (ha->mr.fw_hbt_en)) {
			fw_heart_beat = RD_REG_DWORD(&reg->fwheartbeat);
			if (fw_heart_beat != ha->mr.old_fw_hbt_cnt) {
				ha->mr.old_fw_hbt_cnt = fw_heart_beat;
				ha->mr.fw_hbt_miss_cnt = 0;
			} else {
				ha->mr.fw_hbt_miss_cnt++;
				if (ha->mr.fw_hbt_miss_cnt ==
				    QLAFX00_HEARTBEAT_MISS_CNT) {
					set_bit(ISP_ABORT_NEEDED,
					    &ha->dpc_flags);
					qla2xxx_wake_dpc(ha);
					ha->mr.fw_hbt_miss_cnt = 0;
				}
			}
		}
		ha->mr.fw_hbt_cnt = QLAFX00_HEARTBEAT_INTERVAL;
	}

	if (test_bit(FX00_RESET_RECOVERY, &ha->dpc_flags)) {
		/* Reset recovery to be performed in timer routine */
		aenmbx0 = RD_REG_DWORD(&reg->aenmailbox0);
		if (ha->mr.fw_reset_timer_exp) {
			set_bit(ISP_ABORT_NEEDED, &ha->dpc_flags);
			qla2xxx_wake_dpc(ha);
			ha->mr.fw_reset_timer_exp = 0;
		}
		if (aenmbx0 == MBA_FW_RESTART_CMPLT) {
			/* Wake up DPC to rescan the targets */
			set_bit(FX00_TARGET_SCAN, &ha->dpc_flags);
			clear_bit(FX00_RESET_RECOVERY, &ha->dpc_flags);
			qla2xxx_wake_dpc(ha);
			ha->mr.fw_reset_timer_tick = QLAFX00_RESET_INTERVAL;
		} else if ((aenmbx0 == MBA_FW_STARTING) &&
		    (!ha->mr.fw_hbt_en)) {
			ha->mr.fw_hbt_en = 1;
		} else if (!ha->mr.fw_reset_timer_tick) {
			if (aenmbx0 == ha->mr.old_aenmbx0_state)
				ha->mr.fw_reset_timer_exp = 1;
			ha->mr.fw_reset_timer_tick = QLAFX00_RESET_INTERVAL;
		} else if (aenmbx0 == 0xFFFFFFFF) {
			uint32_t data0, data1;

			DEBUG2(qla_printk(KERN_INFO, ha,
			    "Reinit QLAFX00_PEX0_WIN0_BASE_ADDR_REG\n"));
			data0 = QLAFX00_RD_REG(ha,
			    QLAFX00_BAR1_BASE_ADDR_REG);
			data1 = QLAFX00_RD_REG(ha,
			    QLAFX00_PEX0_WIN0_BASE_ADDR_REG);

			data0 &= 0xffff0000;
			data1 &= 0x0000ffff;

			QLAFX00_WR_REG(ha,
			    QLAFX00_PEX0_WIN0_BASE_ADDR_REG,
			    (data0 | data1));
		} else if ((aenmbx0 & 0xFF00) == MBA_FW_POLL_STATE) {
			ha->mr.fw_reset_timer_tick =
			    QLAFX00_MAX_RESET_INTERVAL;
		} else if (aenmbx0 == MBA_FW_RESET_FCT) {
			ha->mr.fw_reset_timer_tick =
			    QLAFX00_MAX_RESET_INTERVAL;
		}
		ha->mr.old_aenmbx0_state = aenmbx0;
		ha->mr.fw_reset_timer_tick--;
	}
	if (test_bit(FX00_CRITEMP_RECOVERY, &ha->dpc_flags)) {
		/*
		 * Critical temperature recovery to be
		 * performed in timer routine
		 */
		if (ha->mr.fw_critemp_timer_tick == 0) {
			tempc = QLAFX00_GET_TEMPERATURE(ha);
			DEBUG2(qla_printk(KERN_INFO, ha,
			    "ISPFx00(%s): Critical temp timer, "
			    "current SOC temperature: %d.\n",
			    __func__, tempc));
			if (tempc < ha->mr.critical_temperature) {
				set_bit(ISP_ABORT_NEEDED, &ha->dpc_flags);
				clear_bit(FX00_CRITEMP_RECOVERY,
				    &ha->dpc_flags);
				qla2xxx_wake_dpc(ha);
			}
			ha->mr.fw_critemp_timer_tick =
			    QLAFX00_CRITEMP_INTERVAL;
		} else {
			ha->mr.fw_critemp_timer_tick--;
		}
	}

}

/*
 *  qlfx00a_reset_initialize
 *      Re-initialize after a iSA device reset.
 *
 * Input:
 *      ha  = adapter block pointer.
 *
 * Returns:
 *      0 = success
 */
int
qlafx00_reset_initialize(scsi_qla_host_t *ha)
{

	if (ha->device_flags & DFLG_DEV_FAILED) {
		DEBUG2(qla_printk(KERN_INFO, ha,
		    "Device in failed state\n"));
		return QLA_SUCCESS;
	}

	ha->flags.mr_reset_hdlr_active = 1;

	if (ha->flags.online) {
		scsi_block_requests(ha->host);
		qlafx00_abort_isp_cleanup(ha, false);
	}

	qla_printk(KERN_INFO, ha,
	    "(%s): succeeded.\n", __func__);
	ha->flags.mr_reset_hdlr_active = 0;
	return QLA_SUCCESS;
}

/*
 *  qlafx00_abort_isp
 *      Resets ISP and aborts all outstanding commands.
 *
 * Input:
 *      ha  = adapter block pointer.
 *
 * Returns:
 *      0 = success
 */
int
qlafx00_abort_isp(scsi_qla_host_t *ha)
{

	if (ha->flags.online) {
		if (unlikely(pci_channel_offline(ha->pdev) &&
		    ha->flags.pci_channel_io_perm_failure)) {
			clear_bit(ISP_ABORT_RETRY, &ha->dpc_flags);
			return QLA_SUCCESS;
		}

		scsi_block_requests(ha->host);
		qlafx00_abort_isp_cleanup(ha, false);
	} else {
		scsi_block_requests(ha->host);
		clear_bit(ISP_ABORT_NEEDED, &ha->dpc_flags);
		ha->qla_stats.total_isp_aborts++;
		ha->isp_ops->reset_chip(ha);
		set_bit(FX00_RESET_RECOVERY, &ha->dpc_flags);
		/* Clear the Interrupts */
		QLAFX00_CLR_INTR_REG(ha, QLAFX00_HST_INT_STS_BITS);
	}


	qla_printk(KERN_INFO, ha,
	    "(%s): succeeded.\n", __func__);
	return QLA_SUCCESS;
}

static inline fc_port_t	*
qlafx00_get_fcport(struct scsi_qla_host *ha, int tgt_id)
{
	fc_port_t	*fcport;

	/* Check for matching device in remote port list. */
	fcport = NULL;
	list_for_each_entry(fcport, &ha->vp_fcports, list) {
		if (fcport->tgt_id == tgt_id) {
			DEBUG2(qla_printk(KERN_INFO, ha,
			    "Matching fcport(%p) found with TGT-ID: 0x%x "
			    "and Remote TGT_ID: 0x%x\n",
			    fcport, fcport->tgt_id, tgt_id));
			break;
		}
	}
	return (fcport);
}

static void
qlafx00_tgt_detach(struct scsi_qla_host *ha, int tgt_id)
{
	fc_port_t	*fcport;

	qla_printk(KERN_INFO, ha,
	    "Detach TGT-ID: 0x%x\n", tgt_id);

	fcport = qlafx00_get_fcport(ha, tgt_id);
	if (!fcport)
		return;

	qla2x00_mark_device_lost(ha, fcport, 0, 0);

	return;
}

/*
 * qlafx00_find_target
 *
 * Input:
 *	ha = adapter block pointer.
 *	dev = database device entry pointer.
 *
 * Returns:
 *	0 = success.
 *
 * Context:
 *	Kernel context.
 */
static int
qlafx00_find_target(scsi_qla_host_t *ha,
	struct list_head *new_fcports, uint32_t	tgt_id)
{
	int		rval;
	fc_port_t	*fcport, *new_fcport;
	int		found;

	rval = QLA_SUCCESS;

	if (test_bit(LOOP_RESYNC_NEEDED, &ha->dpc_flags)) {
		return (rval);
	}

	/* Allocate temporary rmtport for any new rmtports discovered. */
	new_fcport = qla2x00_alloc_fcport(ha, GFP_KERNEL);
	if (new_fcport == NULL) {
		return (QLA_MEMORY_ALLOC_FAILED);
	}

	if (STATE_TRANSITION(ha)) {
		kfree(new_fcport);
		return (rval);
	}

	/* Send get target node info */
	new_fcport->tgt_id = tgt_id;
	rval = qlafx00_fx_disc(ha, new_fcport, FXDISC_GET_TGT_NODE_INFO);
	if (rval != QLA_SUCCESS) {
		kfree(new_fcport);
		return (rval);
	}
	/* Locate matching device in database. */
	found = 0;
	list_for_each_entry(fcport, &ha->vp_fcports, list) {
		if (memcmp(new_fcport->port_name, fcport->port_name,
		    WWN_SIZE)) {
			continue;
		}
		found++;

		/*
		 * If address the same and state FCS_ONLINE, nothing
		 * changed.
		 */
		if (fcport->tgt_id == new_fcport->tgt_id &&
		    atomic_read(&fcport->state) == FCS_ONLINE) {
			break;
		}

		/*
		 * Tgt ID changed or device was marked to be updated.
		 */
		qla_printk(KERN_INFO, ha,
			"TGT-ID changed(%s). Existing tgt id 0x%x state 0x%x "
			"wwnn = %llx wwpn = %llx.\n",
			__func__, fcport->tgt_id,
			atomic_read(&fcport->state),
			(unsigned long long)wwn_to_u64(fcport->node_name),
			(unsigned long long)wwn_to_u64(fcport->port_name));
		qla_printk(KERN_INFO, ha,
			"TGT-ID Announce(%s): Discovered tgt id 0x%x wwnn = %llx "
			"wwpn = %llx.\n",
			__func__, new_fcport->tgt_id,
			(unsigned long long)wwn_to_u64(new_fcport->node_name),
			(unsigned long long)wwn_to_u64(new_fcport->port_name));

		if (atomic_read(&fcport->state) != FCS_ONLINE) {
			fcport->old_tgt_id = fcport->tgt_id;
			fcport->tgt_id = new_fcport->tgt_id;
			qla2x00_update_fcport(ha, fcport);
		} else {
			qla2x00_mark_device_lost(ha, fcport, 0, 0);
			set_bit(LOOP_RESYNC_NEEDED, &ha->dpc_flags);
			kfree(new_fcport);
			return(rval);
		}

		break;
	}

	if (found) {
		kfree(new_fcport);
		return (rval);
	}
	/* If device was not in our fcports list, then add it. */
	list_add_tail(&new_fcport->list, new_fcports);

	return (rval);
}

int
qlafx00_configure_target(struct scsi_qla_host *ha, uint32_t tgtidx)
{
	int	rval;
	fc_port_t	*fcport, *rmptemp;
	LIST_HEAD(new_fcports);

	rval = qlafx00_find_target(ha, &new_fcports, tgtidx);
	if (rval != QLA_SUCCESS) {
		set_bit(LOOP_RESYNC_NEEDED, &ha->dpc_flags);
		return (rval);
	} else if (atomic_read(&ha->loop_down_timer) ||
		    test_bit(LOOP_RESYNC_NEEDED, &ha->dpc_flags)) {
		return (rval);
	}

	/*
	 * Add the new devices to our devices list.
	 */
	list_for_each_entry_safe(fcport, rmptemp, &new_fcports, list) {
		if (test_bit(LOOP_RESYNC_NEEDED, &ha->dpc_flags))
			break;

		qla2x00_update_fcport(ha, fcport);
		list_move_tail(&fcport->list, &ha->vp_fcports);
		qla_printk(KERN_INFO, ha,
		    "Attach new target id 0x%x wwnn = %llx "
		    "wwpn = %llx for fcport: %p.\n", fcport->tgt_id,
		    (unsigned long long)wwn_to_u64(fcport->node_name),
		    (unsigned long long)wwn_to_u64(fcport->port_name), fcport);
	}

	/* Free all new device structures not processed. */
	list_for_each_entry_safe(fcport, rmptemp, &new_fcports, list) {
		list_del(&fcport->list);
		kfree(fcport);
	}
	return (rval);
}

int
qlafx00_process_aen(struct scsi_qla_host *ha, struct qla_work_evt *evt)
{
	int rval = 0;

	switch (evt->u.aenfx.evtcode) {
		case MBA_SYSTEM_ERR:		/* System Error */
		case MBA_REQ_TRANSFER_ERR:	/* Request Transfer Error */
		case MBA_RSP_TRANSFER_ERR:	/* Response Transfer Error */
		    break;
		case MBA_SHUTDOWN_REQUESTED:	/* FW shutdown pending */
			set_bit(ISP_ABORT_NEEDED, &ha->dpc_flags);
			rval = qlafx00_driver_shutdown(ha, 10);
			break;
		case MBA_PORT_UPDATE:		/* Port database update */
			if (evt->u.aenfx.mbx[1] == 0) {
				if (evt->u.aenfx.mbx[2] == 1) {
					if (!ha->flags.fw_tgt_reported)
						ha->flags.fw_tgt_reported = 1;
					atomic_set(&ha->loop_down_timer, 0);
					atomic_set(&ha->loop_state, LOOP_UP);
					set_bit(LOOP_RESYNC_NEEDED,
					    &ha->dpc_flags);
					qla2xxx_wake_dpc(ha);
				} else if (evt->u.aenfx.mbx[2] == 2) {
					qlafx00_tgt_detach(ha, evt->u.aenfx.mbx[3]);
				}
			} else if (evt->u.aenfx.mbx[1] == 0xffff) {
				if (evt->u.aenfx.mbx[2] == 1) {
					if (!ha->flags.fw_tgt_reported)
						ha->flags.fw_tgt_reported = 1;
					set_bit(LOOP_RESYNC_NEEDED, &ha->dpc_flags);
				} else if (evt->u.aenfx.mbx[2] == 2) {
						ha->device_flags |= DFLG_NO_CABLE;
						qla2x00_mark_all_devices_lost(ha, 1);
				}
			}
			break;
		case MBA_FW_RESTART_CMPLT:
			break;
		case MBA_FW_INIT_FAILURE:
			break;
		case QLAFX00_MBA_TEMP_OVER:
		case QLAFX00_MBA_TEMP_CRIT:	/* Critical temparture event */
		    qla_printk(KERN_INFO, ha,
			"Process critical temperature event "
			"aenmb[0]: %x.\n",
			evt->u.aenfx.evtcode);
		    scsi_block_requests(ha->host);
		    qlafx00_abort_isp_cleanup(ha, true);
		    scsi_unblock_requests(ha->host);
		    break;
	}
	return (rval);
}

static void
qlafx00_update_host_attr(scsi_qla_host_t *ha, port_info_data_t *pinfo)
{
	u64 port_name = 0, node_name = 0;

	port_name = (unsigned long long)wwn_to_u64(pinfo->port_name);
	node_name = (unsigned long long)wwn_to_u64(pinfo->node_name);

	fc_host_node_name(ha->host) = node_name;
	fc_host_port_name(ha->host) = port_name;
	if (!pinfo->port_type)
		ha->current_topology = ISP_CFG_F;
	if (pinfo->link_status == QLAFX00_LINK_STATUS_UP)
		atomic_set(&ha->loop_state, LOOP_READY);
	else if (pinfo->link_status == QLAFX00_LINK_STATUS_DOWN)
		atomic_set(&ha->loop_state, LOOP_DOWN);
	ha->link_data_rate = (uint16_t)pinfo->link_config;
}

void
qlafx00_wait_for_fxiocb_completion(scsi_qla_host_t *ha,
    struct qla_fxiocb_rqst *pfxiocb_rqst)
{
	unsigned long timeout;

	if (unlikely(pci_channel_offline(ha->pdev)))
		return;

	timeout = ((ha->r_a_tov / 10 * 2) + 20) * HZ;
	if (pfxiocb_rqst->rqst_type == FX_IOCB_DCMD) {
		if (!wait_for_completion_timeout(&pfxiocb_rqst->u.fxiocb.fxiocb_comp, timeout)) {
			DEBUG2(qla_printk(KERN_WARNING, ha,
			    "FXIOCB DCMD timeout.\n"));
		}
	} else {
		if (!wait_for_completion_timeout(&pfxiocb_rqst->u.iocb_fx00.piocb_comp, timeout)) {
			DEBUG2(qla_printk(KERN_WARNING, ha,
			    "FXIOCB BCMD timeout.\n"));
		}
	}
}

int
qlafx00_fx_disc(scsi_qla_host_t *ha, fc_port_t *fcport, uint8_t fx_type)
{
	struct qla_fxiocb_rqst fxiocb_rqst;
	struct fxiocb *fxio = &fxiocb_rqst.u.fxiocb;
	struct fxdisc_entry_fx00 fx_iocb;
	int rval = QLA_FUNCTION_FAILED;
	unsigned long flags;
	struct host_system_info *phost_info;
	struct register_host_info *preg_hsi;
	struct timeval tv;

	memset(fxio, 0, sizeof(struct fxiocb));
	switch (fx_type) {
	case FXDISC_GET_CONFIG_INFO:
	    fxio->flags = SRB_FXDISC_RESP_DMA_VALID;
	    fxio->rsp_len = sizeof(config_info_data_t);
	    break;
	case FXDISC_GET_PORT_INFO:
	    fxio->flags =
		SRB_FXDISC_RESP_DMA_VALID | SRB_FXDISC_REQ_DWRD_VALID;
	    fxio->rsp_len = QLAFX00_PORT_DATA_INFO;
	    fxio->req_data = fcport->port_id;
	    break;
	case FXDISC_GET_TGT_NODE_INFO:
	    fxio->flags =
		SRB_FXDISC_RESP_DMA_VALID | SRB_FXDISC_REQ_DWRD_VALID;
	    fxio->rsp_len = QLAFX00_TGT_NODE_INFO;
	    fxio->req_data = fcport->tgt_id;
	    break;
	case FXDISC_GET_TGT_NODE_LIST:
	    fxio->flags =
		SRB_FXDISC_RESP_DMA_VALID | SRB_FXDISC_REQ_DWRD_VALID;
	    fxio->rsp_len = QLAFX00_TGT_NODE_LIST_SIZE;
	    break;
	case FXDISC_REG_HOST_INFO:
	    fxio->flags = SRB_FXDISC_REQ_DMA_VALID;
	    fxio->req_len = sizeof(struct register_host_info);
	    break;
	default:
	    break;
	}
	fxio->req_func_type = fx_type;

	if (fxio->flags & SRB_FXDISC_REQ_DMA_VALID) {
		fxio->req_addr = dma_alloc_coherent(&ha->pdev->dev,
		    fxio->req_len,
		    &fxio->req_dma_handle, GFP_KERNEL);
		if (!fxio->req_addr) {
			goto done;
		}

		if (fx_type == FXDISC_REG_HOST_INFO) {
			preg_hsi = (struct register_host_info *)
				fxio->req_addr;
			phost_info = &preg_hsi->hsi;
			memset(preg_hsi, 0, sizeof(struct register_host_info));
			phost_info->os_type = OS_TYPE_LINUX;
			strncpy(phost_info->sysname,
			    system_utsname.sysname, SYSNAME_LENGTH);
			strncpy(phost_info->nodename,
			    system_utsname.nodename, NODENAME_LENGTH);
			strncpy(phost_info->release,
			    system_utsname.release, RELEASE_LENGTH);
			strncpy(phost_info->version,
			    system_utsname.version, VERSION_LENGTH);
			strncpy(phost_info->machine,
			    system_utsname.machine, MACHINE_LENGTH);
			strncpy(phost_info->domainname,
			    system_utsname.domainname, DOMNAME_LENGTH);
			strncpy(phost_info->hostdriver,
			    QLA2XXX_VERSION, VERSION_LENGTH);
			do_gettimeofday(&tv);
			preg_hsi->utc = (uint64_t)tv.tv_sec;
			DEBUG2(qla_printk(KERN_INFO, ha,
			    "ISP%04X: Host registartion with firmware\n",
			    ha->pdev->device));
			DEBUG2(qla_printk(KERN_INFO,ha,
			    "os_type = '%d', sysname = '%s', nodname = '%s'\n",
			    phost_info->os_type,
			    phost_info->sysname,
			    phost_info->nodename));
			DEBUG2(qla_printk(KERN_INFO,ha,
			    "release = '%s', version = '%s'\n",
			    phost_info->release,
			    phost_info->version));
			DEBUG2(qla_printk(KERN_INFO,ha,
			    "machine = '%s' "
			    "domainname = '%s', hostdriver = '%s'\n",
			    phost_info->machine,
			    phost_info->domainname,
			    phost_info->hostdriver));
			DEBUG12(qla2x00_dump_buffer((uint8_t *)phost_info,
			    sizeof(host_system_info_t)));
		}
	}

	if (fxio->flags & SRB_FXDISC_RESP_DMA_VALID) {
		fxio->rsp_addr = dma_alloc_coherent(&ha->pdev->dev,
		    fxio->rsp_len,
		    &fxio->rsp_dma_handle, GFP_KERNEL);
		if (!fxio->rsp_addr) {
			goto done_unmap_req;
		}
	}

	fxiocb_rqst.rqst_type = FX_IOCB_DCMD;
	init_completion(&fxio->fxiocb_comp);
	if (qlafx00_fxdisc_iocb(fcport, &fxiocb_rqst, &fx_iocb)) {
		qla_printk(KERN_WARNING, ha,
		    "Not enough room in req_q, retry later\n");
		goto done_unmap_rsp;
	}

	spin_lock_irqsave(&ha->hardware_lock, flags);
	qla2x00_isp_cmd(ha);
	spin_unlock_irqrestore(&ha->hardware_lock, flags);

	qlafx00_wait_for_fxiocb_completion(ha, &fxiocb_rqst);


	if (fx_type == FXDISC_GET_CONFIG_INFO) {
		config_info_data_t *pinfo =
		    (config_info_data_t *) fxio->rsp_addr;
		memcpy(&ha->mr.product_name, pinfo->product_name,
		    sizeof(ha->mr.product_name));
		memcpy(&ha->mr.symbolic_name, pinfo->symbolic_name,
		    sizeof(ha->mr.symbolic_name));
		memcpy(&ha->mr.serial_num, pinfo->serial_num,
		    sizeof(ha->mr.serial_num));
		memcpy(&ha->mr.hw_version, pinfo->hw_version,
		    sizeof(ha->mr.hw_version));
		memcpy(&ha->mr.fw_version, pinfo->fw_version,
		    sizeof(ha->mr.fw_version));
		qla_strim(ha->mr.fw_version);
		memcpy(&ha->mr.uboot_version, pinfo->uboot_version,
		    sizeof(ha->mr.uboot_version));
		memcpy(&ha->mr.fru_serial_num, pinfo->fru_serial_num,
		    sizeof(ha->mr.fru_serial_num));
		ha->mr.critical_temperature = pinfo->nominal_temp_value;
		ha->mr.extended_io_enabled = (pinfo->enabled_capabilities &
		    QLAFX00_EXTENDED_IO_EN_MASK) != 0;
	} else if (fx_type == FXDISC_GET_PORT_INFO) {
		port_info_data_t *pinfo =
		    (port_info_data_t *) fxio->rsp_addr;
		memcpy(ha->node_name, pinfo->node_name, WWN_SIZE);
		memcpy(ha->port_name, pinfo->port_name, WWN_SIZE);
		ha->d_id.b.domain = pinfo->port_id[0];
		ha->d_id.b.area = pinfo->port_id[1];
		ha->d_id.b.al_pa = pinfo->port_id[2];
		qlafx00_update_host_attr(ha, pinfo);
		DEBUG12(qla_printk(KERN_INFO, ha,
		    "Dumping Port Info: (%s).\n", __func__));
		DEBUG12(qla2x00_dump_buffer((uint8_t *)pinfo, 16));
		qla_printk(KERN_INFO, ha,
		    "Dumping Port Info: (%s).\n", __func__);
		qla2x00_dump_buffer((uint8_t *)pinfo, 16);
	} else if (fx_type == FXDISC_GET_TGT_NODE_INFO) {
		qlafx00_tgt_node_info_t *pinfo =
		    (qlafx00_tgt_node_info_t *) fxio->rsp_addr;
		memcpy(fcport->node_name, pinfo->tgt_node_wwnn, WWN_SIZE);
		memcpy(fcport->port_name, pinfo->tgt_node_wwpn, WWN_SIZE);
		fcport->port_type = FCT_TARGET;
		DEBUG12(qla_printk(KERN_INFO, ha,
		    "Dumping Target Node info: (%s).\n", __func__));
		DEBUG12(qla2x00_dump_buffer((uint8_t *)pinfo, 16));
	} else if (fx_type == FXDISC_GET_TGT_NODE_LIST) {
		qlafx00_tgt_node_info_t *pinfo =
		    (qlafx00_tgt_node_info_t *) fxio->rsp_addr;
		DEBUG12(qla_printk(KERN_INFO, ha,
		    "Dumping Target Node List info: (%s).\n", __func__));
		DEBUG12(qla2x00_dump_buffer((uint8_t *)pinfo, 16));
		memcpy(ha->gid_list, pinfo, QLAFX00_TGT_NODE_LIST_SIZE);
	}
	rval = fxio->result;

done_unmap_rsp:
	if (fxio->flags & SRB_FXDISC_RESP_DMA_VALID) {
		DEBUG2(qla_printk(KERN_INFO, ha,
		    "(%s): Releasing resp dma address\n", __func__));
		dma_free_coherent(&ha->pdev->dev, fxio->rsp_len,
		    fxio->rsp_addr, fxio->rsp_dma_handle);
	}

done_unmap_req:
	if (fxio->flags & SRB_FXDISC_REQ_DMA_VALID) {
		DEBUG2(qla_printk(KERN_INFO, ha,
		    "(%s): Releasing req dma address\n", __func__));
		dma_free_coherent(&ha->pdev->dev, fxio->req_len,
		    fxio->req_addr, fxio->req_dma_handle);
	}
done:
	return rval;
}

int
qlafx00_async_abt_cmd(uint32_t abt_handle, srb_t *cmd_sp)
{
	scsi_qla_host_t *ha = cmd_sp->ha;
	fc_port_t *fcport = cmd_sp->fcport;
	struct abort_iocb_entry_fx00 abt_iocb;
	struct abt abort;
	unsigned long flags;
	uint32_t rval = CS_TIMEOUT;

	DEBUG2(qla_printk(KERN_INFO, ha,
	    "Abort command issued - hdl=%x, target_id=%x\n",
	    abt_handle, fcport->tgt_id));

	abort.cmd_hndl = abt_handle;

	init_completion(&abort.abt_comp);
	qlafx00_abort_iocb(fcport, &abt_iocb, &abort);
	spin_lock_irqsave(&ha->hardware_lock, flags);
	qla2x00_isp_cmd(ha);
	spin_unlock_irqrestore(&ha->hardware_lock, flags);

	qlafx00_wait_for_abort_completion(ha, &abort);

	rval = abort.comp_status == CS_COMPLETE ?
	    QLA_SUCCESS: QLA_FUNCTION_FAILED;

	return rval;
}

/*
* qlafx00_initialize_adapter
*      Initialize board.
*
* Input:
*      ha = adapter block pointer.
*
* Returns:
*      0 = success
*/
int
qlafx00_initialize_adapter(scsi_qla_host_t *ha)
{
	int	rval;
	uint32_t tempc;

	/* Clear adapter flags. */
	ha->flags.online = 0;
	ha->flags.chip_reset_done = 0;
	ha->flags.reset_active = 0;
	ha->flags.pci_channel_io_perm_failure = 0;
	ha->flags.eeh_busy = 0;
	atomic_set(&ha->loop_down_timer, LOOP_DOWN_TIME);
	atomic_set(&ha->loop_state, LOOP_DOWN);
	ha->device_flags = DFLG_NO_CABLE;
	ha->dpc_flags = 0;
	ha->flags.management_server_logged_in = 0;
	ha->marker_needed = 0;
	ha->mbx_flags = 0;
	ha->isp_abort_cnt = 0;
	ha->beacon_blink_led = 0;

	qla_printk(KERN_INFO, ha, "Configuring PCI space...\n");
	rval = ha->isp_ops->pci_config(ha);
	if (rval) {
		printk("scsi(%ld): Unable to configure PCI space.\n",
		    ha->host_no);
		return (rval);
	}

	rval = qlafx00_init_fw_ready(ha);
	if ( rval != QLA_SUCCESS) {
		return rval;
	}

	qlafx00_save_queue_ptrs(ha);

	rval = qlafx00_config_queues(ha);
	if (rval != QLA_SUCCESS)
		return rval;

	rval = qla2x00_init_rings(ha);
	ha->flags.chip_reset_done = 1;
	ha->isp_ops->disable_intrs(ha);

	tempc = QLAFX00_GET_TEMPERATURE(ha);
	DEBUG2(qla_printk(KERN_INFO, ha,
	    "ISPFX00(%s): Critical temp timer, current SOC temperature: 0x%x\n",
	    __func__, tempc));

	return (rval);
}

uint32_t
qlafx00_fw_state_show(struct class_device *cdev, char *buf)
{
	int rval = QLA_FUNCTION_FAILED;
	uint32_t state[1];
	scsi_qla_host_t *ha = to_qla_host(class_to_shost(cdev));
	scsi_qla_host_t *pha = to_qla_parent(ha);

	if (qla2x00_reset_active(pha))
		qla_printk(KERN_WARNING, pha,
		    "ISP reset active.\n");
	else if (!pha->flags.eeh_busy) {
		rval = qlafx00_get_firmware_state(pha, state);
	}
	if (rval != QLA_SUCCESS)
		memset(state, -1, sizeof(state));

	return state[0];
}

void
qlafx00_get_host_speed(struct Scsi_Host *shost)
{
	struct scsi_qla_host *ha = (struct scsi_qla_host *)
		(shost_priv(shost));
	u32 speed = FC_PORTSPEED_UNKNOWN;

	switch (ha->link_data_rate) {
	case QLAFX00_PORT_SPEED_2G:
		speed = FC_PORTSPEED_2GBIT;
		break;
	case QLAFX00_PORT_SPEED_4G:
		speed = FC_PORTSPEED_4GBIT;
		break;
	case QLAFX00_PORT_SPEED_8G:
		speed = FC_PORTSPEED_8GBIT;
		break;
	case QLAFX00_PORT_SPEED_10G:
		speed = FC_PORTSPEED_10GBIT;
		break;
	}
	fc_host_speed(shost) = speed;
}

static inline cont_a64_entry_t *
qlafx00_prep_cont_type1_iocb(scsi_qla_host_t *ha,
    cont_a64_entry_t *lcont_pkt)
{
	cont_a64_entry_t *cont_pkt;

	/* Adjust ring index. */
	ha->req_ring_index++;
	if (ha->req_ring_index == ha->request_q_length) {
		ha->req_ring_index = 0;
		ha->request_ring_ptr = ha->request_ring;
	} else {
		ha->request_ring_ptr++;
	}

	cont_pkt = (cont_a64_entry_t *)ha->request_ring_ptr;

	/* Load packet defaults. */
	*((uint32_t *)(&lcont_pkt->entry_type)) =
	    __constant_cpu_to_le32(CONTINUE_A64_TYPE_FX00);

	return (cont_pkt);
}

static inline void
qlafx00_build_scsi_iocbs(srb_t *sp, cmd_type_7_fx00_t *cmd_pkt,
    uint16_t tot_dsds, cmd_type_7_fx00_t *lcmd_pkt)
{
	uint16_t	avail_dsds;
	uint32_t	*cur_dsd;
	scsi_qla_host_t	*ha;
	struct scsi_cmnd *cmd;
	int cont;
	cont_a64_entry_t lcont_pkt;
	cont_a64_entry_t *cont_pkt;

	cmd = sp->cmd;
	cont = 0;
	cont_pkt = NULL;

	/* Update entry type to indicate Command Type 3 IOCB */
	*((uint32_t *)(&lcmd_pkt->entry_type)) =
	    __constant_cpu_to_le32(FX00_COMMAND_TYPE_7);

	/* No data transfer */
	if (!scsi_bufflen(cmd) || cmd->sc_data_direction == DMA_NONE) {
		lcmd_pkt->byte_count = __constant_cpu_to_le32(0);
		return;
	}

	ha = sp->ha;

	/* Set transfer direction */
	if (cmd->sc_data_direction == DMA_TO_DEVICE) {
		lcmd_pkt->cntrl_flags =
		    __constant_cpu_to_le16(TMF_WRITE_DATA);
		ha->qla_stats.output_bytes += scsi_bufflen(cmd);
	} else if (cmd->sc_data_direction == DMA_FROM_DEVICE) {
		lcmd_pkt->cntrl_flags =
		    __constant_cpu_to_le16(TMF_READ_DATA);
		ha->qla_stats.input_bytes += scsi_bufflen(cmd);
	}

	/* One DSD is available in the Command Type 3 IOCB */
	avail_dsds = 1;
	cur_dsd = (uint32_t *)&lcmd_pkt->dseg_0_address;

	/* Load data segments */
	if (scsi_sg_count(cmd) != 0) {
		struct	scatterlist *cur_seg;
		struct	scatterlist *end_seg;

		cur_seg = scsi_sglist(cmd);
		end_seg = cur_seg + tot_dsds;
		while (cur_seg < end_seg) {
			dma_addr_t	sle_dma;

			/* Allocate additional continuation packets? */
			if (avail_dsds == 0) {
				/*
				 * Five DSDs are available in the Continuation
				 * Type 1 IOCB.
				 */
				memset(&lcont_pkt, 0, REQUEST_ENTRY_SIZE);
				cont_pkt = qlafx00_prep_cont_type1_iocb(ha,
				    &lcont_pkt);
				cur_dsd = (uint32_t *)lcont_pkt.dseg_0_address;
				avail_dsds = 5;
				cont = 1;
			}

			sle_dma = sg_dma_address(cur_seg);
			*cur_dsd++ = cpu_to_le32(LSD(sle_dma));
			*cur_dsd++ = cpu_to_le32(MSD(sle_dma));
			*cur_dsd++ = cpu_to_le32(sg_dma_len(cur_seg));
			avail_dsds--;

			cur_seg++;
			if (avail_dsds == 0 && cont == 1) {
				cont = 0;
				memcpy_toio((void __iomem *)cont_pkt,
				    &lcont_pkt, REQUEST_ENTRY_SIZE);
				DEBUG19(qla_printk(KERN_INFO, ha,
				    "(%s): Dumping Cont(1) IOCB buffer.\n", __func__));
				DEBUG19(qla2x00_dump_buffer(
				    (uint8_t *)&lcont_pkt, REQUEST_ENTRY_SIZE));
			}
		}
	}
	if (avail_dsds != 0 && cont == 1) {
		memcpy_toio((void __iomem *)cont_pkt, &lcont_pkt,
		    REQUEST_ENTRY_SIZE);
		DEBUG19(qla_printk(KERN_INFO, ha,
		    "(%s): Dumping Cont(2) IOCB buffer.\n", __func__));
		DEBUG19(qla2x00_dump_buffer(
		    (uint8_t *)cont_pkt, REQUEST_ENTRY_SIZE));
	}
}

/**
 * qlafx00_start_scsi() - Send a SCSI command to the ISP
 * @sp: command to send to the ISP
 *
 * Returns non-zero if a failure occured, else zero.
 */
int
qlafx00_start_scsi(srb_t *sp)
{
	int		ret;
	unsigned long   flags;
	scsi_qla_host_t	*ha;
	struct scsi_cmnd *cmd;
	uint32_t        index;
	uint32_t	handle;
	struct scatterlist *sg;
	uint16_t	cnt;
	uint16_t	req_cnt;
	uint16_t	tot_dsds;
	device_reg_t __iomem *reg = ISP_QUE_REG(sp->ha);
	cmd_type_7_fx00_t *cmd_pkt;
	cmd_type_7_fx00_t lcmd_pkt;
	struct scsi_lun llun;
	char		tag[2];

	/* Setup device pointers. */
	ret = 0;
	ha = sp->ha;

	cmd = sp->cmd;
	/* So we know we haven't pci_map'ed anything yet */
	tot_dsds = 0;


	// Forcing marker needed for now
	ha->marker_needed = 0;

	if (ha->marker_needed != 0) {
		if (qla2x00_marker(ha, 0, 0, MK_SYNC_ALL) != QLA_SUCCESS) {
			return QLA_FUNCTION_FAILED;
		}
		ha->marker_needed = 0;
	}

	/* Acquire ring specific lock */
	spin_lock_irqsave(&ha->hardware_lock, flags);

	/* Check for room in outstanding command list. */
	handle = ha->current_outstanding_cmd;
	for (index = 1; index < MAX_OUTSTANDING_COMMANDS; index++) {
		handle++;
		if (handle == MAX_OUTSTANDING_COMMANDS)
			handle = 1;
		if (!ha->outstanding_cmds[handle])
			break;
	}
	if (index == MAX_OUTSTANDING_COMMANDS)
		goto queuing_error;

	/* Map the sg table so we have an accurate count of sg entries needed */
	if (scsi_sg_count(cmd)) {
		sg = scsi_sglist(cmd);
		tot_dsds = pci_map_sg(ha->pdev, sg, scsi_sg_count(cmd),
		    cmd->sc_data_direction);
		if (tot_dsds == 0)
			goto queuing_error;
	} else if (scsi_bufflen(cmd)) {
		dma_addr_t      req_dma;

		req_dma = pci_map_single(ha->pdev, cmd->request_buffer,
		    scsi_bufflen(cmd), cmd->sc_data_direction);
		if (dma_mapping_error(req_dma))
			goto queuing_error;

		sp->dma_handle = req_dma;
		tot_dsds = 1;
	}

	req_cnt = qla24xx_calc_iocbs(tot_dsds);
	if (ha->req_q_cnt < (req_cnt + 2)) {
		cnt = (uint16_t)RD_REG_DWORD_RELAXED(&reg->ispfx00.req_q_out);
		if (ha->req_ring_index < cnt)
			ha->req_q_cnt = cnt - ha->req_ring_index;
		else
			ha->req_q_cnt = ha->request_q_length -
				(ha->req_ring_index - cnt);
		if (ha->req_q_cnt < (req_cnt + 2))
			goto queuing_error;
	}

	/* Build command packet. */
	ha->current_outstanding_cmd = handle;
	ha->outstanding_cmds[handle] = sp;
	sp->ha = ha;
	sp->cmd->host_scribble = (unsigned char *)(unsigned long)handle;
	ha->req_q_cnt -= req_cnt;

	cmd_pkt = (struct cmd_type_7_fx00 *)ha->request_ring_ptr;

	memset(&lcmd_pkt, 0, REQUEST_ENTRY_SIZE);

	lcmd_pkt.handle = handle;
	lcmd_pkt.handle_hi = 0;
	lcmd_pkt.dseg_count = cpu_to_le16(tot_dsds);
	lcmd_pkt.tgt_idx = cpu_to_le16(sp->fcport->tgt_id);

	int_to_scsilun(sp->cmd->device->lun, &llun);
	host_to_adap((uint8_t *)&llun, (uint8_t *)&lcmd_pkt.lun,
	    sizeof(lcmd_pkt.lun));

	/* Update tagged queuing modifier -- default is TSK_SIMPLE (0). */
	if (scsi_populate_tag_msg(cmd, tag)) {
		switch (tag[0]) {
		case HEAD_OF_QUEUE_TAG:
			lcmd_pkt.task = TSK_HEAD_OF_QUEUE;
			break;
		case ORDERED_QUEUE_TAG:
			lcmd_pkt.task = TSK_ORDERED;
			break;
		}
	}

	/* Load SCSI command packet. */
	host_to_adap(cmd->cmnd, lcmd_pkt.fcp_cdb, sizeof(lcmd_pkt.fcp_cdb));
	lcmd_pkt.byte_count = cpu_to_le32((uint32_t)scsi_bufflen(cmd));

	/* Build IOCB segments */
	qlafx00_build_scsi_iocbs(sp, cmd_pkt, tot_dsds, &lcmd_pkt);

	/* Set total data segment count. */
	lcmd_pkt.entry_count = (uint8_t)req_cnt;

	DEBUG19(qla_printk(KERN_INFO, ha,
	    "(%s): Dumping IOCB buffer.\n", __func__));
	DEBUG19(qla2x00_dump_buffer(
	    (uint8_t *)&lcmd_pkt, REQUEST_ENTRY_SIZE));

	memcpy_toio((void __iomem *)cmd_pkt, &lcmd_pkt, REQUEST_ENTRY_SIZE);
	wmb();

	/* Adjust ring index. */
	ha->req_ring_index++;
	if (ha->req_ring_index == ha->request_q_length) {
		ha->req_ring_index = 0;
		ha->request_ring_ptr = ha->request_ring;
	} else
		ha->request_ring_ptr++;

	sp->flags |= SRB_DMA_VALID;

	/* Set chip new ring index. */
	WRT_REG_DWORD(&reg->ispfx00.req_q_in, ha->req_ring_index);
	QLAFX00_SET_HST_INTR(ha, ha->rqstq_intr_code);

	spin_unlock_irqrestore(&ha->hardware_lock, flags);
	return QLA_SUCCESS;

queuing_error:
	if (scsi_sg_count(cmd) && tot_dsds) {
		sg = scsi_sglist(cmd);
		pci_unmap_sg(ha->pdev, sg, scsi_sg_count(cmd),
		    cmd->sc_data_direction);
	} else if (tot_dsds) {
		pci_unmap_single(ha->pdev, sp->dma_handle,
		    scsi_bufflen(cmd), cmd->sc_data_direction);
	}
	spin_unlock_irqrestore(&ha->hardware_lock, flags);

	return QLA_FUNCTION_FAILED;
}

uint32_t
qlafx00_tm_iocb(fc_port_t *fcport, struct tsk_mgmt_entry_fx00 *tskm_iocb, struct
    tmf *ptmf)
{
	scsi_qla_host_t *pha = to_qla_parent(fcport->ha);
	uint32_t handle, index;
	int rval = QLA_FUNCTION_FAILED;
	struct tsk_mgmt_entry_fx00 *ptskm_iocb = NULL;
	unsigned long flags;
	struct scsi_lun llun;
	srb_t *sp;

	sp = mempool_alloc(pha->srb_mempool, GFP_ATOMIC);
	if (!sp)
		return rval;

	memset(sp, 0, sizeof(srb_t));
	spin_lock_irqsave(&pha->hardware_lock, flags);

	/* Check for room in outstanding command list. */
	handle = pha->current_outstanding_cmd;
	for (index = 1; index < MAX_OUTSTANDING_COMMANDS; index++) {
		handle++;
		if (handle == MAX_OUTSTANDING_COMMANDS)
			handle = 1;
		if (!pha->outstanding_cmds[handle])
			break;
	}
	if (index == MAX_OUTSTANDING_COMMANDS)
		return rval;

        ptskm_iocb = (struct tsk_mgmt_entry_fx00 *)qla2x00_req_pkt(pha);
        if (ptskm_iocb == NULL) {
                DEBUG2_3(printk("%s(%ld): failed to allocate FX IOCB.\n",
                    __func__, pha->host_no));

                return rval;
        }

	/* Prep command array. */
	sp->flags |= SRB_FX00_TMF;
	sp->ctx = (void *)ptmf;
	pha->current_outstanding_cmd = handle;
	pha->outstanding_cmds[handle] = sp;
	spin_unlock_irqrestore(&pha->hardware_lock, flags);

	memset(ptskm_iocb, 0, sizeof(struct tsk_mgmt_entry_fx00));
	tskm_iocb->entry_type = TSK_MGMT_IOCB_TYPE_FX00;
        tskm_iocb->entry_count = 1;
	tskm_iocb->handle = cpu_to_le32(handle);
	tskm_iocb->handle_hi = 0;
	tskm_iocb->timeout = ((pha->r_a_tov / 10 * 2) +5);
	tskm_iocb->tgt_id = cpu_to_le16(fcport->tgt_id);
	tskm_iocb->control_flags = cpu_to_le32(ptmf->flags);
	if (tskm_iocb->control_flags == TCF_LUN_RESET) {
		int_to_scsilun(ptmf->lun, &llun);
		host_to_adap((uint8_t *)&llun, (uint8_t *)&tskm_iocb->lun,
			sizeof(struct scsi_lun));
	}
	DEBUG19(qla_printk(KERN_WARNING, pha,
	    "Dumping tsk_mgmt_entry_fx00 buffer.\n"));
	DEBUG19(qla2x00_dump_buffer(
	    (uint8_t *)tskm_iocb, sizeof(struct tsk_mgmt_entry_fx00)));

	memcpy((void __iomem *)ptskm_iocb, tskm_iocb,
	    sizeof(struct tsk_mgmt_entry_fx00));
	wmb();

	return QLA_SUCCESS;
}

uint32_t
qlafx00_abort_iocb(fc_port_t *fcport, struct abort_iocb_entry_fx00 *abt_iocb,
		struct abt *pabt)
{
	scsi_qla_host_t *pha = to_qla_parent(fcport->ha);
	uint32_t handle, index;
	int rval = QLA_FUNCTION_FAILED;
	struct abort_iocb_entry_fx00 *pabt_iocb = NULL;
	unsigned long flags;
	srb_t *sp;

	sp = mempool_alloc(pha->srb_mempool, GFP_ATOMIC);
	if (!sp)
		return rval;


	spin_lock_irqsave(&pha->hardware_lock, flags);

	/* Check for room in outstanding command list. */
	handle = pha->current_outstanding_cmd;
	for (index = 1; index < MAX_OUTSTANDING_COMMANDS; index++) {
		handle++;
		if (handle == MAX_OUTSTANDING_COMMANDS)
			handle = 1;
		if (!pha->outstanding_cmds[handle])
			break;
	}
	if (index == MAX_OUTSTANDING_COMMANDS)
		return rval;

        pabt_iocb = (struct abort_iocb_entry_fx00*)qla2x00_req_pkt(pha);
        if (pabt_iocb == NULL) {
                DEBUG2_3(printk("%s(%ld): failed to allocate FX IOCB.\n",
                    __func__, pha->host_no));

                return rval;
        }

	/* Prep command array. */
	pha->current_outstanding_cmd = handle;
	sp->ctx = (void *)pabt_iocb;
	sp->flags |= SRB_FX00_ABT;
	pha->outstanding_cmds[handle] = sp;
	spin_unlock_irqrestore(&pha->hardware_lock, flags);

	memset(abt_iocb, 0, sizeof(struct abort_iocb_entry_fx00));
	abt_iocb->entry_type = ABORT_IOCB_TYPE_FX00;;
        abt_iocb->entry_count = 1;
        abt_iocb->handle = cpu_to_le32(handle);
	abt_iocb->abort_handle = cpu_to_le32(pabt->cmd_hndl);
	abt_iocb->tgt_id_sts = cpu_to_le16(fcport->tgt_id);
	memcpy((void __iomem *)pabt_iocb, abt_iocb,
	    sizeof(struct abort_iocb_entry_fx00));
	wmb();

	return QLA_SUCCESS;

}

uint32_t
qlafx00_fxdisc_iocb(fc_port_t *fcport, struct qla_fxiocb_rqst *pfxiocb_rqst,
    struct fxdisc_entry_fx00 *fx_iocb)
{
	struct fxdisc_entry_fx00 *pfx_iocb = NULL;
	scsi_qla_host_t *pha = to_qla_parent(fcport->ha);
	uint32_t handle, index;
	int rval = QLA_FUNCTION_FAILED;
	fxiocb_t *fxio;
	qla_mt_iocb_fx00_t *piocb_rqst;
	struct qla_fx00_pt_req *pfx00_pt_req;
	unsigned long flags;
	srb_t *sp;

	sp = mempool_alloc(pha->srb_mempool, GFP_ATOMIC);
	if (!sp)
		return rval;

	memset(sp, 0, sizeof(srb_t));

	spin_lock_irqsave(&pha->hardware_lock, flags);

	/* Check for room in outstanding command list. */
	handle = pha->current_outstanding_cmd;
	for (index = 1; index < MAX_OUTSTANDING_COMMANDS; index++) {
		handle++;
		if (handle == MAX_OUTSTANDING_COMMANDS)
			handle = 1;
		if (!pha->outstanding_cmds[handle])
			break;
	}
	if (index == MAX_OUTSTANDING_COMMANDS)
		return rval;

        pfx_iocb = (struct fxdisc_entry_fx00 *)qla2x00_req_pkt(pha);
        if (pfx_iocb == NULL) {
                DEBUG2_3(printk("%s(%ld): failed to allocate FX IOCB.\n",
                    __func__, pha->host_no));

                return rval;
        }

	/* Prep command array. */
	pha->current_outstanding_cmd = handle;
	sp->ctx = (void *)pfxiocb_rqst;
	sp->flags |= SRB_FX00_IOCTL;
	pha->outstanding_cmds[handle] = sp;
	spin_unlock_irqrestore(&pha->hardware_lock, flags);

	memset(fx_iocb, 0, sizeof(struct fxdisc_entry_fx00));
	fx_iocb->entry_type = FX00_IOCB_TYPE;
        fx_iocb->entry_count = 1;
        fx_iocb->handle = cpu_to_le32(handle);

	if (pfxiocb_rqst->rqst_type == FX_IOCB_DCMD) {
		fxio = &pfxiocb_rqst->u.fxiocb;
		fx_iocb->func_num =
		    cpu_to_le16(fxio->req_func_type);
		fx_iocb->adapid = cpu_to_le32(fxio->adapter_id);
		fx_iocb->adapid_hi = cpu_to_le32(fxio->adapter_id_hi);
		fx_iocb->reserved_0 = cpu_to_le32(fxio->reserved_0);
		fx_iocb->reserved_1 = cpu_to_le32(fxio->reserved_1);
		fx_iocb->dataword_extra = cpu_to_le32(fxio->req_data_extra);

		if (fxio->flags & SRB_FXDISC_REQ_DMA_VALID) {
			fx_iocb->req_dsdcnt = cpu_to_le16(1);
			fx_iocb->req_xfrcnt =
			    cpu_to_le16(fxio->req_len);
			fx_iocb->dseg_rq_address[0] =
			    cpu_to_le32(LSD(fxio->req_dma_handle));
			fx_iocb->dseg_rq_address[1] =
			    cpu_to_le32(MSD(fxio->req_dma_handle));
			fx_iocb->dseg_rq_len =
			    cpu_to_le32(fxio->req_len);
		}

		if (fxio->flags & SRB_FXDISC_RESP_DMA_VALID) {
			fx_iocb->rsp_dsdcnt = cpu_to_le16(1);
			fx_iocb->rsp_xfrcnt =
			    cpu_to_le16(fxio->rsp_len);
			fx_iocb->dseg_rsp_address[0] =
			    cpu_to_le32(LSD(fxio->rsp_dma_handle));
			fx_iocb->dseg_rsp_address[1] =
			    cpu_to_le32(MSD(fxio->rsp_dma_handle));
			fx_iocb->dseg_rsp_len =
			    cpu_to_le32(fxio->rsp_len);
		}

		if (fxio->flags & SRB_FXDISC_REQ_DWRD_VALID) {
			fx_iocb->dataword =
			    cpu_to_le32(fxio->req_data);
		}
		fx_iocb->flags = fxio->flags;
	} else {
		piocb_rqst = &pfxiocb_rqst->u.iocb_fx00;
		pfx00_pt_req = (struct qla_fx00_pt_req *)&piocb_rqst->fx00_req;

		fx_iocb->func_num = cpu_to_le16(pfx00_pt_req->fw_func);
		fx_iocb->adapid = cpu_to_le32(pfx00_pt_req->fw_adapter_id);
		fx_iocb->adapid_hi =
		    cpu_to_le32(pfx00_pt_req->fw_adapter_id_hi);
		fx_iocb->reserved_0 = cpu_to_le32(pfx00_pt_req->fw_reserved1);
		fx_iocb->reserved_1 = pfx00_pt_req->fw_reserved2;
		fx_iocb->dataword = cpu_to_le32(pfx00_pt_req->fw_ioctl_data);
		fx_iocb->dataword_extra =
		    cpu_to_le32(pfx00_pt_req->fw_ioctl_data_ext);

		if (pfx00_pt_req->fw_flags & SRB_FXDISC_REQ_DMA_VALID) {
			fx_iocb->req_dsdcnt = cpu_to_le16(1);
			fx_iocb->req_xfrcnt =
			    cpu_to_le16(pfx00_pt_req->request_len);
			fx_iocb->dseg_rq_address[0] =
			    cpu_to_le32(LSD(piocb_rqst->fx00_req_dma));
			fx_iocb->dseg_rq_address[1] =
			    cpu_to_le32(MSD(piocb_rqst->fx00_req_dma));
			fx_iocb->dseg_rq_len =
			    cpu_to_le32(pfx00_pt_req->request_len);
		}

		if (pfx00_pt_req->fw_flags & SRB_FXDISC_RESP_DMA_VALID) {
			fx_iocb->rsp_dsdcnt = cpu_to_le16(1);
			fx_iocb->rsp_xfrcnt =
			    cpu_to_le16(pfx00_pt_req->response_len);
			fx_iocb->dseg_rsp_address[0] =
			    cpu_to_le32(LSD(piocb_rqst->fx00_resp_dma));
			fx_iocb->dseg_rsp_address[1] =
			    cpu_to_le32(MSD(piocb_rqst->fx00_resp_dma));
			fx_iocb->dseg_rsp_len =
			    cpu_to_le32(pfx00_pt_req->response_len);
		}
	}

	DEBUG19(qla_printk(KERN_WARNING, pha,
	    "Dumping fxdisc_entry_fx00 buffer.\n"));
	DEBUG19(qla2x00_dump_buffer(
	    (uint8_t *)fx_iocb, sizeof(struct fxdisc_entry_fx00)));

	memcpy((void __iomem *)pfx_iocb, fx_iocb,
	    sizeof(struct fxdisc_entry_fx00));
	wmb();

	return QLA_SUCCESS;

}

/**
 * qlafx00_mbx_completion() - Process mailbox command completions.
 * @ha: SCSI driver HA context
 * @mb0: Mailbox0 register
 */
static void
qlafx00_mbx_completion(scsi_qla_host_t *ha, uint16_t mb0)
{
	uint16_t	cnt;
	uint16_t __iomem *wptr;
	struct device_reg_fx00 __iomem *reg = &ha->iobase->ispfx00;

	if (!ha->mcp32)
		DEBUG2_3(printk("%s(%ld): MBX pointer ERROR\n", __func__,
		    ha->host_no));

	/* Load return mailbox registers. */
	ha->flags.mbox_int = 1;
	ha->mailbox_out32[0] = mb0;
	wptr = (uint16_t __iomem *)&reg->mailbox17;

	for (cnt = 1; cnt < ha->mbx_count; cnt++) {
		ha->mailbox_out32[cnt] = RD_REG_WORD(wptr);
		wptr++;
	}
}

/**
 * qlafx00_async_event() - Process aynchronous events.
 * @ha: SCSI driver HA context
 */
void
qlafx00_async_event(scsi_qla_host_t *ha)
{
	struct device_reg_fx00 __iomem *reg = &ha->iobase->ispfx00;

	ha->aenmb[1] = RD_REG_DWORD(&reg->aenmailbox1);
	ha->aenmb[2] = RD_REG_DWORD(&reg->aenmailbox2);
	ha->aenmb[3] = RD_REG_DWORD(&reg->aenmailbox3);
	qla2xxx_log_aen(ha, ha->aenmb[0], ha->aenmb[1], ha->aenmb[2],
	    ha->aenmb[3]);

	/* Setup to process RIO completion. */
	switch (ha->aenmb[0]) {
	case QLAFX00_MBA_SYSTEM_ERR:		/* System Error */
		qla_printk(KERN_WARNING, ha,
		    "ISP System Error - mbx1=%x.\n", ha->aenmb[0]);
		set_bit(ISP_ABORT_NEEDED, &ha->dpc_flags);
		qla2xxx_wake_dpc(ha);
		break;

	case QLAFX00_MBA_SHUTDOWN_RQSTD:	/* Shutdown requested */
		DEBUG2(qla_printk(KERN_WARNING, ha,
		    "Asynchronous FW shutdown requested.\n"));
		set_bit(ISP_ABORT_NEEDED, &ha->dpc_flags);
		qla2xxx_wake_dpc(ha);
		break;

	case QLAFX00_MBA_PORT_UPDATE:
		DEBUG2(printk(KERN_INFO
		    "scsi(%ld): Asynchronous PORT UPDATE "
		    "(%x %x %x).\n", ha->host_no, ha->aenmb[1],
		    ha->aenmb[2], ha->aenmb[3]));
		qlafx00_post_aenfx_work(ha, ha->aenmb[0],
		    (uint32_t *)ha->aenmb, 4);
		break;
	case QLAFX00_MBA_TEMP_OVER:	/* Over temperature event */
	case QLAFX00_MBA_TEMP_CRIT:	/* Critical temperature event */
	    qla_printk(KERN_WARNING, ha,
		"Asynchronous critical temperature event received "
		"aenmb[0]: %x\n", ha->aenmb[0]);
	    qlafx00_post_aenfx_work(ha, ha->aenmb[0],
		(uint32_t *)ha->aenmb, 1);
	    break;

	default:
		ha->aenmb[4] = RD_REG_DWORD(&reg->aenmailbox4);
		ha->aenmb[5] = RD_REG_DWORD(&reg->aenmailbox5);
		ha->aenmb[6] = RD_REG_DWORD(&reg->aenmailbox6);
		ha->aenmb[7] = RD_REG_DWORD(&reg->aenmailbox7);
	    DEBUG19(printk("scsi(%ld): AEN: %04x %04X %04x %04x %04x %04x %04x\n",
		ha->host_no, ha->aenmb[1], ha->aenmb[2], ha->aenmb[3],
		ha->aenmb[4], ha->aenmb[5], ha->aenmb[6], ha->aenmb[7]));
	}

}

static inline void
qlafx00_handle_sense(srb_t *sp, uint8_t *sense_data, uint32_t par_sense_len,
    uint32_t sense_len)
{
	struct scsi_cmnd *cp = sp->cmd;

	sp->fw_sense_length = sense_len;
	if (sense_len >= SCSI_SENSE_BUFFERSIZE)
		sense_len = SCSI_SENSE_BUFFERSIZE;

	CMD_ACTUAL_SNSLEN(cp) = sense_len;
	sp->request_sense_length = sense_len;
	sp->request_sense_ptr = cp->sense_buffer;
	if (sp->request_sense_length > par_sense_len)
		sense_len = par_sense_len;

	memcpy(cp->sense_buffer, sense_data, sense_len);

	sp->request_sense_ptr += sense_len;
	sp->request_sense_length -= sense_len;
	sp->fw_sense_length -= sense_len;
	if (sp->fw_sense_length != 0)
		sp->ha->status_srb = sp;

	DEBUG19(printk("%s(): Check condition Sense data, scsi(%ld:%d:%d:%d) "
	    "cmd=%p pid=%ld\n", __func__, sp->ha->host_no, cp->device->channel,
	    cp->device->id, cp->device->lun, cp, cp->serial_number));
	if (sense_len)
		DEBUG5(qla2x00_dump_buffer(cp->sense_buffer,
		    CMD_ACTUAL_SNSLEN(cp)));
}

/**
 * qlafx00_status_entry() - Process a Status IOCB entry.
 * @ha: SCSI driver HA context
 * @pkt: Entry pointer
 */
static void
qlafx00_status_entry(scsi_qla_host_t *ha, void *pkt)
{
	srb_t		*sp;
	fc_port_t	*fcport;
	struct scsi_cmnd *cp;
	struct sts_entry_fx00 *sts;
	uint16_t	comp_status;
	uint16_t	scsi_status;
	uint16_t	ox_id;
	uint8_t		lscsi_status;
	int32_t		resid;
	uint32_t	sense_len, par_sense_len, rsp_info_len, resid_len;
	uint32_t	fw_resid_len;
	uint8_t		*rsp_info = NULL, *sense_data = NULL;
	int logit = 1;
	struct tmf *ptmf;

	sts = (struct sts_entry_fx00 *) pkt;
	comp_status = le16_to_cpu(sts->comp_status);
	scsi_status = le16_to_cpu(sts->scsi_status) & SS_MASK;

	/* Validate handle. */
	if (sts->handle < MAX_OUTSTANDING_COMMANDS) {
		sp = ha->outstanding_cmds[sts->handle];
	} else
		sp = NULL;

	if (sp == NULL) {
		qla_printk(KERN_WARNING, ha,
		    "scsi(%ld): %s Invalid status handle (0x%x).\n",
		    ha->host_no, __func__, sts->handle);
		set_bit(ISP_ABORT_NEEDED, &ha->dpc_flags);
		qla2xxx_wake_dpc(ha);
		return;
	}

	DEBUG19(qla_printk(KERN_INFO, ha,
	    "scsi(%ld): scsi_status: 0x%x, comp_status: 0x%x.\n", ha->host_no,
	    scsi_status, comp_status));

	if (sp->flags & SRB_FX00_TMF) {
		ptmf = (struct tmf *)sp->ctx;
		complete(&ptmf->tmf_comp);
		mempool_free(sp, ha->srb_mempool);
		ha->outstanding_cmds[sts->handle] = NULL;
		return;
	}

	/* Fast path completion. */
	if (comp_status == CS_COMPLETE && scsi_status == 0) {
		qla2x00_process_completed_request(ha, sts->handle);
		return;
	}

	ha->outstanding_cmds[sts->handle] = NULL;
	cp = sp->cmd;
	if (cp == NULL) {
		qla_printk(KERN_WARNING, ha,
		    "scsi(%ld): Command already returned (0x%x/%p).\n",
		    ha->host_no, sts->handle, sp);

		return;
	}

	lscsi_status = scsi_status & STATUS_MASK;
	CMD_ENTRY_STATUS(cp) = sts->entry_status;
	CMD_COMPL_STATUS(cp) = comp_status;
	CMD_SCSI_STATUS(cp) = scsi_status;

	fcport = sp->fcport;

	ox_id = 0;
	sense_len = par_sense_len = rsp_info_len = resid_len = fw_resid_len = 0;
	if (scsi_status & SS_SENSE_LEN_VALID)
		sense_len = le32_to_cpu(sts->sense_len);
	if (scsi_status & (SS_RESIDUAL_UNDER | SS_RESIDUAL_OVER))
		resid_len = le32_to_cpu(sts->residual_len);
	if (comp_status == CS_DATA_UNDERRUN)
		fw_resid_len = le32_to_cpu(sts->residual_len);
	rsp_info = sts->data;
	sense_data = sts->data;
	par_sense_len = sizeof(sts->data);

//TODO
// The rsp_info_len field is not passsed up from FW -- FEATURE ENHANCEMENT??
#if 0
	/* Check for any FCP transport errors. */
	if (scsi_status & SS_RESPONSE_INFO_LEN_VALID) {
		/* Sense data lies beyond any FCP RESPONSE data. */
		if (IS_FWI2_CAPABLE(ha)) {
			sense_data += rsp_info_len;
			par_sense_len -= rsp_info_len;
		}
		if (rsp_info_len > 3 && rsp_info[3]) {
			DEBUG2(qla_printk(KERN_INFO, ha,
			    "scsi(%ld:%d:%d): FCP I/O protocol failure "
			    "(0x%x/0x%x).\n", ha->host_no, cp->device->id,
			    cp->device->lun, rsp_info_len, rsp_info[3]));

			cp->result = DID_BUS_BUSY << 16;
			goto out;
		}
	}
#endif

	/* Check for overrun. */
	if (comp_status == CS_COMPLETE &&
	    scsi_status & SS_RESIDUAL_OVER)
		comp_status = CS_DATA_OVERRUN;

	/*
	 * Based on Host and scsi status generate status code for Linux
	 */
	switch (comp_status) {
	case CS_COMPLETE:
	case CS_QUEUE_FULL:
		if (scsi_status == 0) {
			cp->result = DID_OK << 16;
			break;
		}
		if (scsi_status & (SS_RESIDUAL_UNDER | SS_RESIDUAL_OVER)) {
			resid = resid_len;
			scsi_set_resid(cp, resid);
			CMD_RESID_LEN(cp) = resid;

			if (!lscsi_status &&
			    ((unsigned)(scsi_bufflen(cp) - resid) <
			     cp->underflow)) {
				qla_printk(KERN_INFO, ha,
				    "scsi(%ld:%d:%d): Mid-layer underflow "
				    "detected (0x%x of 0x%x bytes).\n",
				    ha->host_no, cp->device->id,
				    cp->device->lun, resid,
				    scsi_bufflen(cp));

				cp->result = DID_ERROR << 16;
				break;
			}
		}
		cp->result = DID_OK << 16 | lscsi_status;

		if (lscsi_status == SAM_STAT_TASK_SET_FULL) {
			DEBUG2(qla_printk(KERN_INFO, ha,
			    "scsi(%ld:%d:%d) QUEUE FULL detected.\n",
			    ha->host_no, cp->device->id, cp->device->lun));

			/* Adjust queue depth for all luns on the port. */
			if (!ql2xqfulltracking)
				break;
			fcport->last_queue_full = jiffies;
			starget_for_each_device(cp->device->sdev_target,
			    fcport, qla2x00_adjust_sdev_qdepth_down);
			break;
		}
		logit = 0;
		if (lscsi_status != SS_CHECK_CONDITION)
			break;

		memset(cp->sense_buffer, 0, sizeof(cp->sense_buffer));
		if (!(scsi_status & SS_SENSE_LEN_VALID))
			break;

		qlafx00_handle_sense(sp, sense_data, par_sense_len, sense_len);
		break;

	case CS_DATA_UNDERRUN:
		/* Use F/W calculated residual length. */
		resid = fw_resid_len;
		scsi_set_resid(cp, resid);
		if (scsi_status & SS_RESIDUAL_UNDER) {
			CMD_RESID_LEN(cp) = resid;
			if (fw_resid_len != resid_len) {
				DEBUG2(qla_printk(KERN_INFO, ha,
				    "scsi(%ld:%d:%d) Dropped frame(s) detected "
				    "(0x%x of 0x%x bytes), residual length "
				    "mismatch.\n", ha->host_no, cp->device->id,
				    cp->device->lun, resid, scsi_bufflen(cp)));

				cp->result = DID_ERROR << 16 | lscsi_status;
				goto check_scsi_status;
			}

			if (!lscsi_status &&
			    ((unsigned)(scsi_bufflen(cp) - resid) <
			    cp->underflow)) {
				qla_printk(KERN_INFO, ha,
				    "scsi(%ld:%d:%d): Mid-layer underflow "
				    "detected (0x%x of 0x%x bytes).\n",
				     ha->host_no, cp->device->id,
				     cp->device->lun, resid, scsi_bufflen(cp));

				cp->result = DID_ERROR << 16;
				break;
			}
		} else if (lscsi_status != SAM_STAT_TASK_SET_FULL &&
			    lscsi_status != SAM_STAT_BUSY) {
			/* scsi status of task set and busy are considered to be
			 * task no completed.
			 */
			DEBUG2(qla_printk(KERN_INFO, ha,
			    "scsi(%ld:%d:%d) Dropped frame(s) detected (0x%x "
			    "of 0x%x bytes), firmware reported underrun.\n",
			    ha->host_no, cp->device->id, cp->device->lun,
			    resid, scsi_bufflen(cp)));

			cp->result = DID_ERROR << 16 | lscsi_status;
			goto check_scsi_status;
		} else {
			DEBUG2(qla_printk(KERN_INFO, ha,
			    "scsi(%ld:%d:%d) scsi_status: 0x%x, lscsi_status: "
			    "0x%x\n", ha->host_no, cp->device->id,
			    cp->device->lun, scsi_status, lscsi_status));
		}

		cp->result = DID_OK << 16 | lscsi_status;
		logit = 0;

check_scsi_status:
		/*
		 * Check to see if SCSI Status is non zero. If so report SCSI
		 * Status.
		 */
		if (lscsi_status != 0) {

			if (lscsi_status == SAM_STAT_TASK_SET_FULL) {
				DEBUG2(qla_printk(KERN_INFO, ha,
				    "scsi(%ld:%d:%d) QUEUE FULL detected.\n",
				    ha->host_no, cp->device->id,
				    cp->device->lun));
				logit = 1;

				/*
				 * Adjust queue depth for all luns on the
				 * port.
				 */
				if (!ql2xqfulltracking)
					break;

				fcport->last_queue_full = jiffies;
				starget_for_each_device(
				    cp->device->sdev_target, fcport,
				    qla2x00_adjust_sdev_qdepth_down);
				break;
			}
			if (lscsi_status != SS_CHECK_CONDITION)
				break;

			memset(cp->sense_buffer, 0, sizeof(cp->sense_buffer));
			if (!(scsi_status & SS_SENSE_LEN_VALID))
				break;

			qlafx00_handle_sense(sp, sense_data, par_sense_len,
			    sense_len);
		}
		break;

	case CS_PORT_LOGGED_OUT:
	case CS_PORT_CONFIG_CHG:
	case CS_PORT_BUSY:
	case CS_INCOMPLETE:
	case CS_PORT_UNAVAILABLE:
	case CS_TIMEOUT:
	case CS_RESET:

		/*
		 * We are going to have the fc class block the rport
		 * while we try to recover so instruct the mid layer
		 * to requeue until the class decides how to handle this.
		 */
		cp->result = DID_TRANSPORT_DISRUPTED << 16;

		DEBUG2(qla_printk(KERN_INFO, ha,
		    "scsi(%ld:%d:%d) Port down status: port-state=0x%x\n",
		     ha->host_no, cp->device->id, cp->device->lun,
		     atomic_read(&fcport->state)));

		if (atomic_read(&fcport->state) == FCS_ONLINE) {
			qla2x00_mark_device_lost(ha, fcport, 1, 1);
		}
		break;

	case CS_ABORTED:
		cp->result = DID_RESET << 16;
		break;

	default:
		cp->result = DID_ERROR << 16;
		break;
	}

//out:
	if (logit)
		DEBUG2(qla_printk(KERN_INFO, ha,
		    "scsi(%ld:%d:%d) FCP command status: 0x%x-0x%x (0x%x) "
		    "portid=%02x%02x%02x oxid=0x%x ser=0x%lx cdb=%02x%02x%02x "
		    "len=0x%x " "rsp_info=0x%x resid=0x%x fw_resid=0x%x\n",
		    ha->host_no, cp->device->id, cp->device->lun, comp_status,
		    scsi_status, cp->result, fcport->d_id.b.domain,
		    fcport->d_id.b.area, fcport->d_id.b.al_pa, ox_id,
		    cp->serial_number, cp->cmnd[0], cp->cmnd[1], cp->cmnd[2],
		    scsi_bufflen(cp), rsp_info_len, resid_len, fw_resid_len));

	if (ha->status_srb == NULL) {
		qla2x00_sp_compl(ha, sp);
	}
}

/**
 * qlafx00_status_cont_entry() - Process a Status Continuations entry.
 * @ha: SCSI driver HA context
 * @pkt: Entry pointer
 *
 * Extended sense data.
 */
static void
qlafx00_status_cont_entry(scsi_qla_host_t *ha, sts_cont_entry_t *pkt)
{
	uint8_t		sense_sz = 0;
	srb_t		*sp = ha->status_srb;
	struct scsi_cmnd *cp;
	if (!sp){
		DEBUG2(printk("%s(): no SP "
		    "sp=%p.\n", __func__, sp));
		qla_printk(KERN_INFO, ha,
		    "SP is NULL: (sp=%p)\n", sp);
		return;
	}

	if (!sp->fw_sense_length){
		DEBUG2(printk("%s(): no FW sense data "
		    "sp=%p.\n", __func__, sp));
		qla_printk(KERN_INFO, ha,
		    "Residual FW sense data is 0: (sp=%p)\n", sp);
		return;
	}

	cp = sp->cmd;
	if (cp == NULL) {
		DEBUG2(printk("%s(): Cmd is NULL, already returned back to OS "
		    "sp=%p.\n", __func__, sp));

		ha->status_srb = NULL;
		return;
	}

	if (sp->request_sense_length == 0) {
		DEBUG2(printk("%s(): Firmware returned data size larger than "
		    "FC transport can handle, discarding the remaining. "
		    "sp=%p.\n", __func__, sp));
	} else {

		if (sp->request_sense_length > sizeof(pkt->data)) {
			sense_sz = sizeof(pkt->data);
		} else {
			sense_sz = sp->request_sense_length;
		}

		/* Move sense data. */
		memcpy(sp->request_sense_ptr, pkt->data, sense_sz);
		DEBUG5(qla2x00_dump_buffer(sp->request_sense_ptr, sense_sz));

		sp->request_sense_ptr += sense_sz;
		sp->request_sense_length -= sense_sz;

	}
	
	sp->fw_sense_length = (sp->fw_sense_length > sizeof(pkt->data)) ?
	    (sp->fw_sense_length - sizeof(pkt->data)) : 0;
	
	/* Place command on done queue. */
	if (sp->fw_sense_length == 0) {
		ha->status_srb = NULL;
		qla2x00_sp_compl(ha, sp);
	}
}

/**
 * qlafx00_multistatus_entry() - Process Multi response queue entries.
 * @ha: SCSI driver HA context
 */
void qlafx00_multistatus_entry(scsi_qla_host_t *ha, void *pkt)
{
	srb_t *sp = NULL;
	struct multi_sts_entry_fx00 *stsmfx;
	uint32_t handle, hindex, handle_count, i;
	uint32_t *handle_ptr;

	stsmfx = (struct multi_sts_entry_fx00 *) pkt;

	handle_count = stsmfx->handle_count;

	if (handle_count > MAX_HANDLE_COUNT) {
		DEBUG2(qla_printk(KERN_INFO, ha,
		    "Invalid handle count (0x%x).\n", handle_count));
		set_bit(ISP_ABORT_NEEDED, &ha->dpc_flags);
		qla2xxx_wake_dpc(ha);
		return;
	}

	handle_ptr = (uint32_t*) &stsmfx->handles[0];

	for (i = 0; i < handle_count; i++) {
		hindex = le32_to_cpu(*handle_ptr);
		handle = LSW(hindex);

		/* Validate handle. */
		if (handle < MAX_OUTSTANDING_COMMANDS)
			sp = ha->outstanding_cmds[handle];
		else
			sp = NULL;

		if (sp == NULL) {
			DEBUG2(qla_printk(KERN_INFO, ha,
			    "%s Invalid status handle (0x%x).\n", __func__,
			    handle));
			set_bit(ISP_ABORT_NEEDED, &ha->dpc_flags);
			qla2xxx_wake_dpc(ha);
			return;
		}
		qla2x00_process_completed_request(ha, handle);
		handle_ptr++;
	}

}
/**
 * qlafx00_error_entry() - Process an error entry.
 * @ha: SCSI driver HA context
 * @pkt: Entry pointer
 */
static void
qlafx00_error_entry(scsi_qla_host_t *ha, struct sts_entry_fx00 *pkt)
{
	srb_t *sp;

#if defined(QL_DEBUG_LEVEL_19)
	if (pkt->entry_status & RF_INV_E_ORDER)
		qla_printk(KERN_ERR, ha, "%s: Invalid Entry Order\n", __func__);
	else if (pkt->entry_status & RF_INV_E_COUNT)
		qla_printk(KERN_ERR, ha, "%s: Invalid Entry Count\n", __func__);
	else if (pkt->entry_status & RF_INV_E_PARAM)
		qla_printk(KERN_ERR, ha,
		    "%s: Invalid Entry Parameter\n", __func__);
	else if (pkt->entry_status & RF_INV_E_TYPE)
		qla_printk(KERN_ERR, ha, "%s: Invalid Entry Type\n", __func__);
	else if (pkt->entry_status & RF_BUSY)
		qla_printk(KERN_ERR, ha, "%s: Busy\n", __func__);
	else
		qla_printk(KERN_ERR, ha, "%s: UNKNOWN flag error\n", __func__);
#endif

	/* Validate handle. */
	if (pkt->handle < MAX_OUTSTANDING_COMMANDS)
		sp = ha->outstanding_cmds[pkt->handle];
	else
		sp = NULL;

	if (sp) {
		/* Free outstanding command slot. */
		ha->outstanding_cmds[pkt->handle] = NULL;

		/* Bad payload or header */
		if (pkt->entry_status &
		    (RF_INV_E_ORDER | RF_INV_E_COUNT |
		     RF_INV_E_PARAM | RF_INV_E_TYPE)) {
			sp->cmd->result = DID_ERROR << 16;
		} else if (pkt->entry_status & RF_BUSY) {
			sp->cmd->result = DID_BUS_BUSY << 16;
		} else {
			sp->cmd->result = DID_ERROR << 16;
		}
		qla2x00_sp_compl(ha, sp);

	} else if (pkt->entry_type == COMMAND_A64_TYPE || pkt->entry_type ==
		COMMAND_TYPE || pkt->entry_type == COMMAND_TYPE_7
		|| pkt->entry_type == COMMAND_TYPE_6) {
			qla_printk(KERN_WARNING, ha, "Error entry -"
				"invalid handle\n");

			set_bit(ISP_ABORT_NEEDED, &ha->dpc_flags);

			qla2xxx_wake_dpc(ha);
	}
}

/**
 * qlafx00_process_response_queue() - Process response queue entries.
 * @ha: SCSI driver HA context
 */
void
qlafx00_process_response_queue(struct scsi_qla_host *ha)
{
	device_reg_t __iomem *reg = ISP_QUE_REG(ha);
	struct sts_entry_fx00 *pkt = NULL;
	uint32_t handle;
	response_t *lptr;
	struct qla_fxiocb_rqst *pfxiocb_rqst = NULL;
	struct qla_fx00_pt_resp *pfx00_pt_resp = NULL;
	struct abt *pabt = NULL;
	struct ioctl_iocb_entry_fx00 *iocb_pkt = NULL;
	srb_t *sp = NULL;

	while (RD_REG_DWORD(&(ha->response_ring_ptr->signature))
	    != RESPONSE_PROCESSED) {
		lptr = ha->response_ring_ptr;
		memcpy_fromio(ha->fx00_rsp_pkt, lptr, sizeof(ha->fx00_rsp_pkt));
		pkt = (struct sts_entry_fx00 *)ha->fx00_rsp_pkt;

		DEBUG19(printk(KERN_INFO
		    "scsi(%ld): ring_ptr: %p, ring_addr: %p, ring_index: 0x%x\n",
		    ha->host_no, lptr, ha->response_ring, ha->rsp_ring_index));

		ha->rsp_ring_index++;
		if (ha->rsp_ring_index == ha->response_q_length) {
			ha->rsp_ring_index = 0;
			ha->response_ring_ptr = ha->response_ring;
		} else {
			ha->response_ring_ptr++;
		}

		DEBUG19(qla_printk(KERN_INFO, ha,
		    "(%s): **** Dumping Response pkt.\n", __func__));
		DEBUG19(qla2x00_dump_buffer(
		    (uint8_t *)pkt, sizeof(ha->fx00_rsp_pkt)));

		if (pkt->entry_status != 0 &&
		    pkt->entry_type != IOCTL_IOSB_TYPE_FX00) {

			qlafx00_error_entry(ha, (struct sts_entry_fx00 *) pkt);
			goto next_iter;
		}

		DEBUG19(printk(KERN_INFO "scsi(%ld): entry_type: 0x%x.\n",
		    ha->host_no, pkt->entry_type));

		switch (pkt->entry_type) {
		case STATUS_TYPE_FX00:
			qlafx00_status_entry(ha, pkt);
			break;
		case STATUS_CONT_TYPE_FX00:
			qlafx00_status_cont_entry(ha, (sts_cont_entry_t *)pkt);
			break;
		case MULTI_STATUS_TYPE_FX00:
			qlafx00_multistatus_entry(ha, pkt);
			break;
		case ABORT_IOCB_TYPE_FX00:
		    handle = pkt->handle;
		    if (handle >= MAX_OUTSTANDING_COMMANDS) {
			    printk(KERN_WARNING
				"scsi(%ld): ABT_IOCB handle out of range %x\n",
				ha->host_no, handle);
		    }
		    sp = ha->outstanding_cmds[handle];
		    if (!sp) {
			    printk(KERN_WARNING
				"scsi(%ld): Received invalid sp for "
				"handle=0x%x.\n", ha->host_no, handle);
			    return;
		    }
		    if (!(sp->flags & SRB_FX00_ABT)) {
			    printk(KERN_WARNING
				"scsi(%ld): Received invalid flag=0x%x for "
				"handle=0x%x.\n", ha->host_no, sp->flags,
				handle);
			    return;
		    }
		    pabt = (struct abt *)sp->ctx;
		    mempool_free(sp, ha->srb_mempool);
		    ha->outstanding_cmds[handle] = NULL;
		    complete(&pabt->abt_comp);
			break;
		case IOCTL_IOSB_TYPE_FX00:
		    handle = pkt->handle;
		    if (handle >= MAX_OUTSTANDING_COMMANDS) {
			    printk(KERN_WARNING
				"scsi(%ld): Handle out of range %x\n",
				ha->host_no, handle);
		    }
		    sp = ha->outstanding_cmds[handle];
		    if (!sp) {
			    printk(KERN_WARNING
				"scsi(%ld): Received invalid sp for "
				"handle=0x%x.\n", ha->host_no, handle);
			    return;
		    }
		    if (!(sp->flags & SRB_FX00_IOCTL)) {
			    printk(KERN_WARNING
				"scsi(%ld): Received invalid flag=0x%x for "
				"handle=0x%x.\n", ha->host_no, sp->flags,
				handle);
			    return;
		    }
		    pfxiocb_rqst = (struct qla_fxiocb_rqst *)sp->ctx;
		    mempool_free(sp, ha->srb_mempool);
		    ha->outstanding_cmds[handle] = NULL;
		    if (pfxiocb_rqst->rqst_type == FX_IOCB_DCMD)
			    complete(&pfxiocb_rqst->u.fxiocb.fxiocb_comp);
		    else {
			    iocb_pkt = (struct ioctl_iocb_entry_fx00 *)pkt;
			    pfx00_pt_resp = (struct qla_fx00_pt_resp *)&pfxiocb_rqst->u.iocb_fx00.fx00_resp;
			    memset(pfx00_pt_resp, 0, sizeof(struct
				qla_fx00_pt_resp));
			    pfx00_pt_resp->fw_st_reserved1 =
				iocb_pkt->reserved_0;
			    pfx00_pt_resp->fw_st_func = iocb_pkt->comp_func_num;
			    pfx00_pt_resp->fw_st_ioctl_flags =
				iocb_pkt->fw_iotcl_flags;
			    pfx00_pt_resp->fw_st_ioctl_data =
				iocb_pkt->dataword_r;
			    pfx00_pt_resp->fw_st_adapter_id = iocb_pkt->adapid;
			    pfx00_pt_resp->fw_st_adapter_id_hi =
				iocb_pkt->adapid_hi;
			    pfx00_pt_resp->fw_st_ioctl_data_ext =
				iocb_pkt->reserved_1;
			    pfx00_pt_resp->fw_st_residual_cnt =
				iocb_pkt->residuallen;
			    pfx00_pt_resp->fw_st_ioctl_stat =
				iocb_pkt->status;
			    pfx00_pt_resp->fw_st_seq_num =
				iocb_pkt->seq_no;
			    memcpy(pfx00_pt_resp->fw_st_reserved3,
				iocb_pkt->reserved_2,
				sizeof(pfx00_pt_resp->fw_st_reserved3));
			    complete(&pfxiocb_rqst->u.iocb_fx00.piocb_comp);
		    }
			break;
		default:
			/* Type Not Supported. */
			DEBUG19(printk(KERN_WARNING
			    "scsi(%ld): Received unknown response pkt type %x "
			    "entry status=%x.\n",
			    ha->host_no, pkt->entry_type, pkt->entry_status));
			break;
		}
next_iter:
		WRT_REG_DWORD(&lptr->signature, RESPONSE_PROCESSED);
		wmb();
	}
	WRT_REG_DWORD(&reg->ispfx00.rsp_q_out, ha->rsp_ring_index);
}
/**
 * qlafx00_intr_handler() - Process interrupts for the ISPFX00.
 * @irq:
 * @dev_id: SCSI driver HA context
 * @regs:
 *
 * Called by system whenever the host adapter generates an interrupt.
 *
 * Returns handled flag.
 */
irqreturn_t
qlafx00_intr_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	scsi_qla_host_t	*ha;
	struct device_reg_fx00 __iomem *reg;
	int		status;
	unsigned long	flags;
	unsigned long	iter;
	uint32_t	stat;
	uint16_t	mb[8];
	uint32_t	clr_intr = 0;

	ha = (scsi_qla_host_t *) dev_id;
	if (!ha) {
		printk(KERN_INFO
		    "%s(): NULL host pointer\n", __func__);
		return IRQ_NONE;
	}

	reg = &ha->iobase->ispfx00;
	status = 0;

	if (unlikely(pci_channel_offline(ha->pdev)))
		return IRQ_HANDLED;

	spin_lock_irqsave(&ha->hardware_lock, flags);
	for (iter = 50; iter--; clr_intr = 0) {
		stat = QLAFX00_RD_INTR_REG(ha);
		if ((stat & QLAFX00_HST_INT_STS_BITS) == 0)
			break;

		switch (stat & QLAFX00_HST_INT_STS_BITS) {
		case QLAFX00_INTR_MB_CMPLT:
		case QLAFX00_INTR_MB_RSP_CMPLT:
		case QLAFX00_INTR_MB_ASYNC_CMPLT:
		case QLAFX00_INTR_ALL_CMPLT:
			mb[0] = RD_REG_WORD(&reg->mailbox16);
			qlafx00_mbx_completion(ha, mb[0]);
			status |= MBX_INTERRUPT;
			clr_intr |= QLAFX00_INTR_MB_CMPLT;
			break;
		case QLAFX00_INTR_ASYNC_CMPLT:
		case QLAFX00_INTR_RSP_ASYNC_CMPLT:
			ha->aenmb[0] = RD_REG_WORD(&reg->aenmailbox0);
			qlafx00_async_event(ha);
			clr_intr |= QLAFX00_INTR_ASYNC_CMPLT;
			break;
		case QLAFX00_INTR_RSP_CMPLT:
			qlafx00_process_response_queue(ha);
			clr_intr |= QLAFX00_INTR_RSP_CMPLT;
			break;
		default:
			DEBUG2(printk("scsi(%ld): Unrecognized interrupt type "
			    "(%d).\n",
			    ha->host_no, stat));
			break;
		}
		QLAFX00_CLR_INTR_REG(ha, clr_intr);
		QLAFX00_RD_INTR_REG(ha);
	}

	qla2x00_handle_mbx_completion(ha, status);
	spin_unlock_irqrestore(&ha->hardware_lock, flags);

	return IRQ_HANDLED;
}

/** QLAFX00 specific Mailbox implementation functions */

/*
 * qlafx00_mailbox_command
 *	Issue mailbox command and waits for completion.
 *
 * Input:
 *	ha = adapter block pointer.
 *	mcp = driver internal mbx struct pointer.
 *
 * Output:
 *	mb[MAX_MAILBOX_REGISTER_COUNT] = returned mailbox data.
 *
 * Returns:
 *	0 : QLA_SUCCESS = cmd performed success
 *	1 : QLA_FUNCTION_FAILED   (error encountered)
 *	6 : QLA_FUNCTION_TIMEOUT (timeout condition encountered)
 *
 * Context:
 *	Kernel context.
 */
static int
qlafx00_mailbox_command(scsi_qla_host_t *pvha, mbx_cmd_32_t *mcp)
{
	int		rval;
	unsigned long    flags = 0;
	device_reg_t __iomem *reg;
	uint8_t		abort_active;
	uint8_t		io_lock_on;
	uint16_t	command = 0;
	uint32_t	*iptr;
	uint32_t __iomem *optr;
	uint32_t	cnt;
	uint32_t	mboxes;
	unsigned long	wait_time;
	scsi_qla_host_t *ha = to_qla_parent(pvha);

	reg = ha->iobase;
	io_lock_on = ha->flags.init_done;

	rval = QLA_SUCCESS;
	abort_active = test_bit(ABORT_ISP_ACTIVE, &ha->dpc_flags);

	DEBUG11(printk("%s(%ld): entered.\n", __func__, pvha->host_no));

	if (ha->flags.pci_channel_io_perm_failure) {
		DEBUG(printk("%s(%ld): Perm failure on EEH, timeout MBX "
			     "Exiting.\n", __func__, ha->host_no));
		return QLA_FUNCTION_TIMEOUT;
	}

	if (ha->device_flags & DFLG_DEV_FAILED) {
		DEBUG(qla_printk(KERN_ERR, ha,
		    "%s(%ld): Device in failed state, timeout MBX "
		     "Exiting.\n", __func__, ha->host_no));
		return QLA_FUNCTION_TIMEOUT;
	}

	if (ha->flags.isp82xx_fw_hung) {
		/* Setting Link-Down error */
		mcp->mb[0] = MBS_LINK_DOWN_ERROR;
		DEBUG(qla_printk(KERN_ERR, ha,
		    "%s(%ld): Firmware in hung state FW hung = %d\n",
		    __func__, ha->host_no, ha->flags.isp82xx_fw_hung));
		goto premature_exit;
	}

	/*
	 * Wait for active mailbox commands to finish by waiting at most tov
	 * seconds. This is to serialize actual issuing of mailbox cmds during
	 * non ISP abort time.
	 */
	if (!wait_for_completion_timeout(&ha->mbx_cmd_comp,
	    mcp->tov * HZ)) {
		/* Timeout occurred. Return error. */
		DEBUG2_3_11(printk("%s(%ld): cmd access timeout, cmd=0x%x, "
			    "Exiting.\n", __func__, ha->host_no, mcp->mb[0]));
		return QLA_FUNCTION_TIMEOUT;
	}

	ha->flags.mbox_busy = 1;
	/* Save mailbox command for debug */
	ha->mcp32 = mcp;

	DEBUG11(printk("scsi(%ld): prepare to issue mbox cmd=0x%x.\n",
	    ha->host_no, mcp->mb[0]));

	spin_lock_irqsave(&ha->hardware_lock, flags);

	/* Load mailbox registers. */
	optr = (uint32_t __iomem *)&reg->ispfx00.mailbox0;

	iptr = mcp->mb;
	command = mcp->mb[0];
	mboxes = mcp->out_mb;

	for (cnt = 0; cnt < ha->mbx_count; cnt++) {
		if (mboxes & BIT_0)
			WRT_REG_WORD(optr, *iptr);

		mboxes >>= 1;
		optr++;
		iptr++;
	}

#if defined(QL_DEBUG_LEVEL_1)
	printk("%s(%ld): Loaded MBX registers (displayed in bytes) = \n",
	    __func__, ha->host_no);
	qla2x00_dump_buffer((uint8_t *)mcp->mb, 16);
	printk("\n");
	qla2x00_dump_buffer(((uint8_t *)mcp->mb + 0x10), 16);
	printk("\n");
	qla2x00_dump_buffer(((uint8_t *)mcp->mb + 0x20), 8);
	printk("\n");
	printk("%s(%ld): I/O address = %p.\n", __func__, ha->host_no, optr);
	qla2x00_dump_regs(ha);
#endif

	/* Issue set host interrupt command to send cmd out. */
	ha->flags.mbox_int = 0;
	clear_bit(MBX_INTERRUPT, &ha->mbx_cmd_flags);

	/* Unlock mbx registers and wait for interrupt */
	DEBUG11(printk("%s(%ld): going to unlock irq & waiting for interrupt. "
	    "jiffies=%lx.\n", __func__, ha->host_no, jiffies));

	/* Wait for mbx cmd completion until timeout */

	if (!abort_active && io_lock_on) {
		set_bit(MBX_INTR_WAIT, &ha->mbx_cmd_flags);

		QLAFX00_SET_HST_INTR(ha, ha->mbx_intr_code);
		spin_unlock_irqrestore(&ha->hardware_lock, flags);

		wait_for_completion_timeout(&ha->mbx_intr_comp, mcp->tov * HZ);
	} else {
		DEBUG3_11(printk("%s(%ld): cmd=%x POLLING MODE.\n", __func__,
		    ha->host_no, command));
		QLAFX00_SET_HST_INTR(ha, ha->mbx_intr_code);
		spin_unlock_irqrestore(&ha->hardware_lock, flags);

		wait_time = jiffies + mcp->tov * HZ; /* wait at most tov secs */

		while (!ha->flags.mbox_int) {
			if (time_after(jiffies, wait_time))
				break;

			/* Check for pending interrupts. */
			qla2x00_poll(ha);

			if (!ha->flags.mbox_int &&
			    !(IS_QLA2200(ha) &&
			    command == MBC_LOAD_RISC_RAM_EXTENDED))
				msleep(10);
		} /* while */
		DEBUG17(qla_printk(KERN_WARNING, ha,
			"Waited %d sec\n",
			(uint)((jiffies - (wait_time - (mcp->tov * HZ)))/HZ)));
	}

	/* Check whether we timed out */
	if (ha->flags.mbox_int) {
		uint32_t *iptr2;

		DEBUG3_11(printk("%s(%ld): cmd %x completed.\n", __func__,
		    ha->host_no, command));
		/* Got interrupt. Clear the flag. */
		ha->flags.mbox_int = 0;
		clear_bit(MBX_INTERRUPT, &ha->mbx_cmd_flags);

		if (ha->mailbox_out32[0] != MBS_COMMAND_COMPLETE)
			rval = QLA_FUNCTION_FAILED;

		/* Load return mailbox registers. */
		iptr2 = mcp->mb;
		iptr = (uint32_t *)&ha->mailbox_out32[0];
		mboxes = mcp->in_mb;
		for (cnt = 0; cnt < ha->mbx_count; cnt++) {
			if (mboxes & BIT_0)
				*iptr2 = *iptr;

			mboxes >>= 1;
			iptr2++;
			iptr++;
		}
	} else {

		rval = QLA_FUNCTION_TIMEOUT;
	}

	ha->flags.mbox_busy = 0;

	/* Clean up */
	ha->mcp32 = NULL;

	if (abort_active || !io_lock_on) {
		DEBUG11(printk("%s(%ld): checking for additional resp "
		    "interrupt.\n", __func__, ha->host_no));

		/* polling mode for non isp_abort commands. */
		qla2x00_poll(ha);
	}

	if (rval == QLA_FUNCTION_TIMEOUT &&
	    mcp->mb[0] != MBC_GEN_SYSTEM_ERROR) {
		if (!io_lock_on || (mcp->flags & IOCTL_CMD) ||
		    ha->flags.eeh_busy) {
			/* not in dpc. schedule it for dpc to take over. */
			DEBUG(printk("%s(%ld): timeout schedule "
			    "isp_abort_needed.\n", __func__, ha->host_no));

			if (!test_bit(ISP_ABORT_NEEDED, &ha->dpc_flags) &&
			    !test_bit(ABORT_ISP_ACTIVE, &ha->dpc_flags) &&
			    !test_bit(ISP_ABORT_RETRY, &ha->dpc_flags)) {

				qla_printk(KERN_WARNING, ha,
				    "Mailbox command timeout occurred, "
				    "cmd=0x%x, mb[0]=0x%x, eeh_busy=0x%x. "
				    "Scheduling ISP abort.\n", command,
				    mcp->mb[0], ha->flags.eeh_busy);
				set_bit(ISP_ABORT_NEEDED, &ha->dpc_flags);
				qla2xxx_wake_dpc(ha);
			}
		} else if (!abort_active) {
			/* call abort directly since we are in the DPC thread */
			DEBUG(printk("%s(%ld): timeout calling abort_isp\n",
			    __func__, ha->host_no));
			if (!test_bit(ISP_ABORT_NEEDED, &ha->dpc_flags) &&
			    !test_bit(ABORT_ISP_ACTIVE, &ha->dpc_flags) &&
			    !test_bit(ISP_ABORT_RETRY, &ha->dpc_flags)) {

				qla_printk(KERN_WARNING, ha,
				    "Mailbox command timeout occurred, cmd=0x%x "
				    "mb[0]=0x%x. Issuing ISP abort.\n",
				    command, mcp->mb[0]);

				set_bit(ABORT_ISP_ACTIVE, &ha->dpc_flags);
				clear_bit(ISP_ABORT_NEEDED, &ha->dpc_flags);
				if (ha->isp_ops->abort_isp(ha)) {
					/* Failed. retry later. */
					set_bit(ISP_ABORT_NEEDED,
					    &ha->dpc_flags);
				}
				clear_bit(ABORT_ISP_ACTIVE, &ha->dpc_flags);
				DEBUG(printk("%s(%ld): finished abort_isp\n",
				    __func__, ha->host_no));
			}
		}
	}

premature_exit:
	/* Allow next mbx cmd to come in. */
	complete(&ha->mbx_cmd_comp);

	if (rval) {
		DEBUG2_3_11(printk("%s(%ld): **** FAILED. mbx0=%x, mbx1=%x, "
		    "mbx2=%x, cmd=%x ****\n", __func__, ha->host_no,
		    mcp->mb[0], mcp->mb[1], mcp->mb[2], command));
	} else {
		DEBUG11(printk("%s(%ld): done.\n", __func__, ha->host_no));
	}

	return rval;
}

/*
 * qlafx00_driver_shutdown
 *	Indicate a driver shutdown to firmware.
 *
 * Input:
 *	ha = adapter block pointer.
 *
 * Returns:
 *	local function return status code.
 *
 * Context:
 *	Kernel context.
 */
int
qlafx00_driver_shutdown(scsi_qla_host_t *ha, int tmo)
{
	int rval;
	mbx_cmd_32_t mc;
	mbx_cmd_32_t *mcp = &mc;

	DEBUG11(qla_printk(KERN_INFO, ha,
	    "Entered %s.\n", __func__));

	mcp->mb[0] = MBC_MR_DRV_SHUTDOWN;
	mcp->out_mb = MBX_0;
	mcp->in_mb = MBX_0;
	if (tmo)
		mcp->tov = tmo;
	else
		mcp->tov = MBX_TOV_SECONDS;
	mcp->flags = 0;
	rval = qlafx00_mailbox_command(ha, mcp);

	if (rval != QLA_SUCCESS) {
		DEBUG2_3_11(qla_printk(KERN_INFO, ha,
		    "Failed=%x.\n", rval));
	} else {
		DEBUG11(qla_printk(KERN_INFO, ha,
		    "Done %s.\n", __func__));
	}

	return rval;
}

/*
 * qlafx00_get_firmware_state
 *	Get adapter firmware state.
 *
 * Input:
 *	ha = adapter block pointer.
 *	TARGET_QUEUE_LOCK must be released.
 *	ADAPTER_STATE_LOCK must be released.
 *
 * Returns:
 *	qla7xxx local function return status code.
 *
 * Context:
 *	Kernel context.
 */
int
qlafx00_get_firmware_state(scsi_qla_host_t *ha, uint32_t *states)
{
	int rval;
	mbx_cmd_32_t mc;
	mbx_cmd_32_t *mcp = &mc;

	DEBUG11(qla_printk(KERN_INFO, ha,
	    "Entered %s.\n", __func__));

	mcp->mb[0] = MBC_GET_FIRMWARE_STATE;
	mcp->out_mb = MBX_0;
	mcp->in_mb = MBX_1|MBX_0;
	mcp->tov = MBX_TOV_SECONDS;
	mcp->flags = 0;
	rval = qlafx00_mailbox_command(ha, mcp);

	/* Return firmware states. */
	states[0] = mcp->mb[1];

	if (rval != QLA_SUCCESS) {
		DEBUG2_3_11(qla_printk(KERN_INFO, ha,
		    "Failed=%x mb[0]=%x.\n", rval, mcp->mb[0]));
	} else {
		DEBUG11(qla_printk(KERN_INFO, ha,
		    "Done %s.\n", __func__));
	}
	return rval;
}

/*
 * qlafx00_init_firmware
 *	Initialize adapter firmware.
 *
 * Input:
 *	ha = adapter block pointer.
 *	dptr = Initialization control block pointer.
 *	size = size of initialization control block.
 *	TARGET_QUEUE_LOCK must be released.
 *	ADAPTER_STATE_LOCK must be released.
 *
 * Returns:
 *	qlafx00 local function return status code.
 *
 * Context:
 *	Kernel context.
 */
int
qlafx00_init_firmware(scsi_qla_host_t *ha, uint16_t size)
{
	int rval;
	mbx_cmd_32_t mc;
	mbx_cmd_32_t *mcp = &mc;

	DEBUG11(qla_printk(KERN_INFO, ha,
	    "Entered %s.\n", __func__));

	mcp->mb[0] = MBC_INITIALIZE_FIRMWARE;

	mcp->mb[1] = 0;
	mcp->mb[2] = MSD(ha->init_cb_dma);
	mcp->mb[3] = LSD(ha->init_cb_dma);

	mcp->out_mb = MBX_3|MBX_2|MBX_1|MBX_0;
	mcp->in_mb = MBX_0;
	mcp->buf_size = size;
	mcp->flags = MBX_DMA_OUT;
	mcp->tov = MBX_TOV_SECONDS;
	rval = qlafx00_mailbox_command(ha, mcp);

	if (rval != QLA_SUCCESS) {
		DEBUG2_3_11(qla_printk(KERN_INFO, ha,
		    "Failed=%x mb[0]=%x.\n", rval, mcp->mb[0]));
	} else {
		DEBUG11(qla_printk(KERN_INFO, ha,
		    "Done %s.\n", __func__));
	}
	return rval;
}

/*
 * qlafx00_mbx_reg_test
 */
int
qlafx00_mbx_reg_test(scsi_qla_host_t *ha)
{
	int rval;
	mbx_cmd_32_t mc;
	mbx_cmd_32_t *mcp = &mc;

	DEBUG11(qla_printk(KERN_INFO, ha,
	    "Entered %s.\n", __func__));

	mcp->mb[0] = MBC_MAILBOX_REGISTER_TEST;
	mcp->mb[1] = 0xAAAA;
	mcp->mb[2] = 0x5555;
	mcp->mb[3] = 0xAA55;
	mcp->mb[4] = 0x55AA;
	mcp->mb[5] = 0xA5A5;
	mcp->mb[6] = 0x5A5A;
	mcp->mb[7] = 0x2525;
	mcp->mb[8] = 0xBBBB;
	mcp->mb[9] = 0x6666;
	mcp->mb[10] = 0xBB66;
	mcp->mb[11] = 0x66BB;
	mcp->mb[12] = 0xB6B6;
	mcp->mb[13] = 0x6B6B;
	mcp->mb[14] = 0x3636;
	mcp->mb[15] = 0xCCCC;

	mcp->out_mb = MBX_15|MBX_14|MBX_13|MBX_12|MBX_11|MBX_10|MBX_9|MBX_8|
			MBX_7|MBX_6|MBX_5|MBX_4|MBX_3|MBX_2|MBX_1|MBX_0;
	mcp->in_mb = MBX_15|MBX_14|MBX_13|MBX_12|MBX_11|MBX_10|MBX_9|MBX_8|
			MBX_7|MBX_6|MBX_5|MBX_4|MBX_3|MBX_2|MBX_1|MBX_0;
	mcp->buf_size = 0;
	mcp->flags = MBX_DMA_OUT;
	mcp->tov = MBX_TOV_SECONDS;
	rval = qlafx00_mailbox_command(ha, mcp);

	if (rval == QLA_SUCCESS) {
		if (mcp->mb[17] != 0xAAAA || mcp->mb[18] != 0x5555 ||
		    mcp->mb[19] != 0xAA55 || mcp->mb[20] != 0x55AA)
			rval = QLA_FUNCTION_FAILED;
		if (mcp->mb[21] != 0xA5A5 || mcp->mb[22] != 0x5A5A ||
		    mcp->mb[23] != 0x2525 || mcp->mb[24] != 0xBBBB)
			rval = QLA_FUNCTION_FAILED;
		if (mcp->mb[25] != 0x6666 || mcp->mb[26] != 0xBB66 ||
		    mcp->mb[27] != 0x66BB || mcp->mb[28] != 0xB6B6)
			rval = QLA_FUNCTION_FAILED;
		if (mcp->mb[29] != 0x6B6B || mcp->mb[30] != 0x3636 ||
		    mcp->mb[31] != 0xCCCC)
			rval = QLA_FUNCTION_FAILED;
	}

	if (rval != QLA_SUCCESS) {
		DEBUG2_3_11(qla_printk(KERN_INFO, ha,
		    "Failed=%x mb[0]=%x.\n", rval, mcp->mb[0]));
	} else {
		DEBUG11(qla_printk(KERN_INFO, ha,
		    "Done %s.\n", __func__));
	}
	return rval;
}

